!__________________________________________________________________________________________________________________________________
!
! Various tools that can be (linked to) other programs.
!
! This code is part of GETSOURCES = Multi-Scale Multi-Wavelength Source Extraction = Alexander Men'shchikov, DAp IRFU CEA Saclay  
!__________________________________________________________________________________________________________________________________
!                 
!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine osystem 
       
     &            ( lsystem )
!__________________________________________________________________________________________________________________________________
!
! This subroutine determines the operating system it runs on - whether it's UNIX-like or OS/2 (DOS)-like.
!__________________________________________________________________________________________________________________________________
!
       implicit none
       logical  lsystem
!__________________________________________________________________________________________________________________________________
!
       lsystem = .false.
       open ( 666, file='/dev/null', status='unknown', err=66 )
       lsystem = .true.
   66  close ( 666 )

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine round1d ( vect, imax, isigfig )
!__________________________________________________________________________________________________________________________________
!
!  written by: David Clarke; modified by Alexander Men'shchikov
!
!  PURPOSE: This subroutine sets the number of significant figures of the elements in the input vector to the specified number.
!
!  INPUT VARIABLES:
!    vect        vector to be "rounded off"
!    imax        dimension of "vect"
!    isigfig     number of significant figures on output
!
!  OUTPUT VARIABLES:
!    vect        rounded off input vector
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       integer       i, imax, isigfig, i1, i2
       real*8        q1, q2, q3, tiny
       real*8        vect(*)
       parameter     ( tiny = 1.0d-99 )
!__________________________________________________________________________________________________________________________________
!
       if (isigfig .le. 0) return
       q1 = 0.0d0
       do i=1,imax
         q1 = dmax1 ( q1, dabs(vect(i)) )
       enddo
       if (q1 .le. tiny) return
       q1 = dlog10 ( q1 )
       q2 = sign ( 0.5d0, q1 )
       i1 = isigfig - idint ( q1 + q2 + 0.5d0 )
       i2 = min0 ( -1, isigfig - 13 )
       q1 = 10.0d0**i1
       q2 = 10.0d0**i2
       do i=1,imax
         q3 = sign  ( q2, vect(i) )
         vect(i) = dnint ( vect(i) * q1 + q3 ) / q1
       enddo

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       function inellipse

     &          ( x, y, dx, dy, x0, y0, amajor, bminor, thepa )
!__________________________________________________________________________________________________________________________________
!
! This subroutine determines whether a given point (x,y) is inside an ellipse centered on (x0,y0).
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       logical       inellipse
       real*8        x, y, x0, y0, amajor, bminor, thepa, radx, rady, angle, dex, dey, sina, cosa, delx, dely, theta
     &             , pi, pio2, pi2, dx, dy, radx2, rady2, rad2, radell2, sinth2
       parameter   ( pi = 3.14159265358979d0, pio2 = pi / 2.0d0, pi2 = pi * 2.0d0 )
!__________________________________________________________________________________________________________________________________
!
       inellipse = .false.
       
       radx = amajor / 2.0d0  !<-- major radius of the current ellipse
       rady = bminor / 2.0d0  !<-- minor radius of the current ellipse
       angle = (90.0d0 - thepa) * (pi / 180.0d0)  !<-- angle of the ellipse's major axis in radians
         
       dex = (x - x0) * dx
       dey = (y - y0) * dy
       sina = sin ( angle )
       cosa = cos ( angle )
       delx = dex * cosa - dey * sina
       dely = dex * sina + dey * cosa
       
       if (dely .gt. 0.0d0) then
         if (delx .eq. 0.0d0) then
           theta = pio2
         else
           if (delx .gt. 0.0d0) then
             theta = atan ( dely / delx )
           else
             theta = pi - atan ( dely / abs ( delx ) )
           endif
         endif
       else
         if (delx .eq. 0.0d0) then
           theta = pi + pio2
         else
           if (delx .gt. 0.0d0) then
             theta = pi2 - atan ( abs ( dely / delx ) )
           else
             theta = pi + atan ( abs ( dely / delx ) )
           endif
         endif
       endif
       
       radx2 = radx**2
       rady2 = rady**2
       rad2  = delx**2 + dely**2
       sinth2 = sin ( theta )**2
       radell2 = radx2 * rady2 / (rady2 + (radx2 - rady2) * sinth2 + 1.0d-30)
       
       if (rad2 .le. radell2) inellipse = .true.

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine findmin
       
     &            ( xn, yn, xk, yk, nx, ny, image, imin, jmin, imagemin )
!__________________________________________________________________________________________________________________________________
!
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       integer       nx, ny, imin, jmin, l, npt, ic, jc
       real*8        xn, yn, xk, yk, dx, dy, dbll, imagemin
       real*8        image(nx,ny)
!__________________________________________________________________________________________________________________________________
!
!!!       imin = nint ( xn )
!!!       jmin = nint ( yn )
!!!       imagemin = image(imin,jmin)

       npt = 3 * nint ( max ( abs ( xk - xn ), abs ( yk - yn ), 1.0d0 ) )
       dx = (xk - xn) / dble ( npt - 1 )
       dy = (yk - yn) / dble ( npt - 1 )
       imagemin = 1.0d+90
       imin = 0
       jmin = 0
       do l=1,npt
         dbll = dble ( l - 1 )
         ic = nint ( xn + dx * dbll )
         jc = nint ( yn + dy * dbll )
         if (image(ic,jc) .lt. imagemin) then
           imin = ic
           jmin = jc
           imagemin = image(imin,jmin)
         endif
       enddo
       if (imin .eq. 0 .or. jmin .eq. 0) then
         write (*,'(/a,4(1pe10.2),i5)') '   FINDMIN: ERROR: Wrong input?: ', xn, yn, xk, yk, npt
         stop 99     
       endif
       
       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine skipcomm 
       
     &            ( iunit, filename, nheader )
!__________________________________________________________________________________________________________________________________
!
!                                     MCM3D, Alexander Men'shchikov, 12/07/2007
!
! Find data in the character line being read from file (by skipping comment lines) and count the number of header comment lines.
!
! INPUT PARAMETERS:
!
!    iunit    - file unit number
!    filename - file name
!
! OUTPUT PARAMETERS:
!
!    nheader  - number of header comment lines in the file
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       integer       iunit, i, lastc, nheader, nhead, lfn
       character*(*) filename
       character*1   ccomm, first
       character*2   clast, clasto
       external      lastc
!__________________________________________________________________________________________________________________________________
!
       nhead = 0
       first = '?'
       clasto = ''
       do
          nhead = nhead + 1
          read (iunit,'(a)', end=98, err=99) clast
          ccomm = clast(1:1)

          if (first .eq. '?' .and. ccomm .ne. ';' .and. ccomm .ne. ':' .and. ccomm .ne. '*' .and. ccomm .ne. '#' .and. 
     &        ccomm .ne. '!') exit

          if ((first .eq. ';' .and. ccomm .ne. ';') .or. (first .eq. ':' .and. ccomm .ne. ':') .or. 
     &        (first .eq. '*' .and. ccomm .ne. '*') .or. (first .eq. '#' .and. ccomm .ne. '#') .or. 
     &        (first .eq. '!' .and. ccomm .ne. '!') .or. clasto .eq. '!~') exit

!!          if (first .ne. '?' .and. first .ne. ccomm) then
!!            write (*,'(/a)') '   SKIPCOMM: ERROR: Comment symbol differs within header: '''//first//''' /= '''//ccomm//'''.'
!!            stop 99
!!          endif
          if (first .eq. '?') first = ccomm
          clasto = clast
       enddo
 98    nheader = nhead - 1
       backspace ( iunit )
       return

! 98    lfn = lastc ( filename )
!       write (*,'(/a)') '   SKIPCOMM: ERROR: No data found in '''// filename(1:lfn) //'''.'
!       stop 99
!!       nheader = nhead
!!       backspace ( iunit )
!!       return

 99    lfn = lastc ( filename )
       write (*,'(/a,i3)') '   SKIPCOMM: ERROR: Trouble reading '''// filename(1:lfn) //''' at line:', i

       stop 99
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine linescolumns 
       
     &            ( iounit, filename, nheader, nlines, ncolumns )
!__________________________________________________________________________________________________________________________________
!
!                                     MCM3D, Alexander Men'shchikov, 20/08/2007
!
! Find actual number of data columns and data lines in the file, based on just the first data line. File is assumed to be
! positioned at the first data line, after all possible header comments. At the end of this routine, the file is positioned back
! at the first data line.
!
! INPUT PARAMETERS:
!
!    iounit   - file unit number
!    filename - file name
!    nheader  - number of header comment lines in the file
!
! OUTPUT PARAMETERS:
!
!    nlines   - number of data lines in the file
!    ncolumns - number of data columns in the file
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       character*(*) filename
       character*20000 inline
       character*1   cc(20000)
       integer       iounit, nheader, nlines
       integer       ncolumns, i, nl, lastc, lfn
!__________________________________________________________________________________________________________________________________
!
       read (iounit,'(a)',end=97,err=98) inline
       do i=1,20000
         cc(i) = ' '
       enddo
                         
! Find number of columns.

       ncolumns = 0
       if (inline(1:1) .ne. ' ') ncolumns = 1
       do i=2,lastc(inline)
         if (inline(i-1:i-1) .eq. ' ' .and. inline(i:i) .ne. ' ') then
           ncolumns = ncolumns + 1
         endif
       enddo

! Find number of lines.

       nl = 0
       do
         nl = nl + 1
         read (iounit,'(a)',end=91,err=98) cc(nl)
       enddo
91     nlines = nl

! Don't count as data lines those which are footer comments (at the end of the file).
                      
       nl = 0
       do i=nlines,nheader,-1
         if (cc(i) .ne. ';' .and. cc(i) .ne. ':' .and. cc(i) .ne. '#' .and. cc(i) .ne. '!' .and. cc(i) .ne. '*') exit
         nl = nl + 1
       enddo
       nlines = nlines - nl
       rewind ( iounit )
       do i=1,nheader
         read (iounit,'(a)') inline
       enddo
       return
 97    lfn = lastc ( filename )
!!       write (*,'(/a)') '   LINESCOL: WARNING: No data found in '''// filename(1:lfn) //'''.'
       nlines = 0
       ncolumns = 0
       rewind ( iounit )
       return
 98    lfn = lastc ( filename )
       write (*,'(/a)') '   LINESCOL: ERROR: Trouble reading ''' // filename(1:lfn) // '''.'

       stop
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine getnumpix
       
     &            ( fname, nx, ny )
!__________________________________________________________________________________________________________________________________
!
! Kept here for compatibility with GETSOURCES v1. Determine numbers of pixels in a FITS image.
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       character*80  record
       character*(*) fname
       integer       unit, status, readwrite, nkeys, blocksize, nspace, lfn, lastc, i, nx, ny
!__________________________________________________________________________________________________________________________________
!
       status = 0
       readwrite = 0
       lfn = lastc ( fname )
       call ftgiou ( unit, status )
       call ftopen ( unit, fname(1:lfn), readwrite, blocksize, status )
       call ftghsp ( unit, nkeys, nspace, status )
       if (status .gt. 0) then
         call printerr ( status )
         stop 99
       endif
       nx = 0
       ny = 0
       do i=1,nkeys
         call ftgrec ( unit, i, record, status )
         if (index ( record, 'NAXIS1' ) .gt. 0) read (record(10:31),*) nx
         if (index ( record, 'NAXIS2' ) .gt. 0) read (record(10:31),*) ny
       enddo
       if (nx .eq. 0 .or. ny .eq. 0) then
         write (*,'(/a)') '   GETNUMPIX: ERROR: Trouble getting image dimensions in ''' // fname(1:lfn) // '''.'
         stop 99
       endif

       return
       end
 
!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine getpixels
       
     &            ( fname, nx, ny, dx, dy )
!__________________________________________________________________________________________________________________________________
!
! Determine numbers of pixels and pixel sizes in a FITS image.
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       logical       lgotdx, lgotdy, lnocheckdx, lnocheckdy
       character*80  record
       character*(*) fname
       integer       unit, status, readwrite, nkeys, blocksize, nspace, lfn, lastc, i, nx, ny
       real*8        dx, dy
!__________________________________________________________________________________________________________________________________
!
       status = 0
       readwrite = 0
       lfn = lastc ( fname )
       call ftgiou ( unit, status )
       call ftopen ( unit, fname(1:lfn), readwrite, blocksize, status )
       call ftghsp ( unit, nkeys, nspace, status )
       if (status .gt. 0) then
         call printerr ( status )
         stop 99
       endif
       nx = 0
       ny = 0
       lnocheckdx = dx .ge. 1.0d99
       lnocheckdy = dy .ge. 1.0d99
       dx = 0.0d0
       dy = 0.0d0
       lgotdx = .false.
       lgotdy = .false.
       do i=1,nkeys
         call ftgrec ( unit, i, record, status )
         if (index ( record, 'NAXIS1' ) .eq. 1) read (record(10:31),*) nx
         if (index ( record, 'NAXIS2' ) .eq. 1) read (record(10:31),*) ny
         if (index ( record, 'CDELT1' ) .eq. 1) then
           read (record(10:31),*) dx
           lgotdx = .true.
         endif
         if (index ( record, 'CDELT2' ) .eq. 1) then
           read (record(10:31),*) dy
           lgotdy = .true.
         endif
       enddo
       if (lgotdx) dx = abs ( dx ) * 3600.0d0
       if (lgotdy) dy = abs ( dy ) * 3600.0d0
       if (nx .eq. 0 .or. ny .eq. 0) then
         write (*,'(/a)') '   GETPIXELS: ERROR: Trouble getting numbers of pixels in ''' // fname(1:lfn) // '''.'
         stop 99
       endif
       if ((.not.lnocheckdx .and. lgotdx .and. dx .lt. 1.0d-99) .or. (.not.lnocheckdy .and. lgotdy .and. dy .lt. 1.0d-99)) then
         write (*,'(/a)') '   GETPIXELS: ERROR: Trouble getting sizes of pixels in ''' // fname(1:lfn) // '''.'
         stop 99
       endif

       return
       end
 
!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine fdelete 
       
     &            ( fname, status )
!__________________________________________________________________________________________________________________________________
!
! A simple little routine to delete a FITS file
!
! May contain fragments from "The FITSIO Cookbook" v1.1 (April 1995) by William D Pence (HEASARC).
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       character*(*) fname
       integer       status, unit, blocksize
!__________________________________________________________________________________________________________________________________
!
! Simply return if status is greater than zero

       if (status .gt. 0) return

! Get an unused Logical Unit Number to use to open the FITS file

       call ftgiou ( unit, status )

! Try to open the file, to see if it exists

       call ftopen ( unit, fname, 1, blocksize, status )

       if (status .eq. 0) then

! File was opened; so now delete it

         call ftdelt ( unit, status )

       elseif (status .eq. 103) then

! File doesn't exist, so just reset status to zero and clear errors

         status = 0
         call ftcmsg
       else

! There was some other error opening the file; delete the file anyway

         status = 0
         call ftcmsg
         call ftdelt ( unit, status )
       endif

! Free the unit number for later reuse

       call ftfiou ( unit, status )

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine printerr 
       
     &            ( status )
!__________________________________________________________________________________________________________________________________
!
! Print out the FITSIO error messages to the user
!
! May contain fragments from "The FITSIO Cookbook" v1.1 (April 1995) by William D Pence (HEASARC).
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       character*30  errtext
       character*80  errmess
       integer       status
!__________________________________________________________________________________________________________________________________
!
! Check if status is OK (no error); if so, simply return

       if (status .le. 0) return

! Get the text string which describes the error

       call ftgerr ( status, errtext )
       write (*,*) '   FITSIO Error Status =',status,': ',errtext

! Read and print out all the error messages on the FITSIO stack

       call ftgmsg ( errmess )
       do while (errmess .ne. ' ')
         write (*,*) errmess
         call ftgmsg ( errmess )
       enddo

       return
       end
                                                                                                                                   
!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine when 
       
     &            ( lunix, ctime, cdate, ndate, iform )
!__________________________________________________________________________________________________________________________________
!
!  OUTPUT VARIABLES:
!
!    ctime       time returned as hh:mm:ss
!    cdate       date returned as specified by "iform"
!    ndate       number of characters set in "cdate"
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       logical       lunix
       character*(*) ctime, cdate
       character*8   ttime, tdate
       character*12  timda
       character*24  udate
       integer*2     hrs, mins, secs, hsecs, year, month, day
       integer       iform, ndate, i, lastc
       external      lastc
!_______
!
! >>>>> UNCOMMENT THE NEXT LINE TO COMPILE ON UNIX:
!_______
!!!notforgnu       external      fdate
       external      fdate
       external      gettim, getdat
!__________________________________________________________________________________________________________________________________
!
! Date and time: "ttime" is in the format hh:mm:ss, "tdate" is in the format mm/dd/yy, "udate" is in a system-dependent format

       udate = ' '
       hrs   = 0
       mins  = 0
       secs  = 0
       hsecs = 0
       year  = 0
       month = 0
       day   = 0

       if (lunix) then
!_______
!
! >>>>> UNCOMMENT THE NEXT LINE TO COMPILE ON UNIX:
!_______
          call fdate ( udate )

          ttime      = udate(12:19)
          tdate(4:5) = udate( 9:10)
          tdate(7:8) = udate(23:24)
          tdate(3:3) = '/'
          tdate(6:6) = '/'
          if (udate(5:7) .eq. 'Jan') tdate(1:2) = '01'
          if (udate(5:7) .eq. 'Feb') tdate(1:2) = '02'
          if (udate(5:7) .eq. 'Mar') tdate(1:2) = '03'
          if (udate(5:7) .eq. 'Apr') tdate(1:2) = '04'
          if (udate(5:7) .eq. 'May') tdate(1:2) = '05'
          if (udate(5:7) .eq. 'Jun') tdate(1:2) = '06'
          if (udate(5:7) .eq. 'Jul') tdate(1:2) = '07'
          if (udate(5:7) .eq. 'Aug') tdate(1:2) = '08'
          if (udate(5:7) .eq. 'Sep') tdate(1:2) = '09'
          if (udate(5:7) .eq. 'Oct') tdate(1:2) = '10'
          if (udate(5:7) .eq. 'Nov') tdate(1:2) = '11'
          if (udate(5:7) .eq. 'Dec') tdate(1:2) = '12'

       else
!_______
!
! >>>>> COMMENT OUT THE NEXT 2 LINES TO COMPILE ON UNIX:
!_______
!!          call gettim ( hrs, mins, secs, hsecs )
!!          call getdat ( year, month, day )

          write (timda, *) hrs
          if (timda(11:11) .eq. ' ') timda(11:11)='0'
          ttime(1:2) = timda(11:12)
          ttime(3:3) = ':'
          write (timda, *) mins
          if (timda(11:11) .eq. ' ') timda(11:11)='0'
          ttime(4:5) = timda(11:12)
          ttime(6:6) = ':'
          write (timda, *) secs
          if (timda(11:11) .eq. ' ') timda(11:11)='0'
          ttime(7:8) = timda(11:12)
          write (timda, *) month
          if (timda(11:11) .eq. ' ') timda(11:11)='0'
          tdate(1:2) = timda(11:12)
          tdate(3:3) = '/'
          write (timda, *) day
          if (timda(11:11) .eq. ' ') timda(11:11)='0'
          tdate(4:5) = timda(11:12)
          tdate(6:6) = '/'
          write (timda, *) year
          tdate(7:8) = timda(11:12)
       endif

! Convert to desired format.

       if (iform .eq. 1) then
         ctime(1:8) = ttime
         cdate(1:8) = tdate
         cdate(1:2) = tdate(4:5)
         cdate(4:5) = tdate(1:2)
         ndate      = 8
       endif

       if (iform .eq. 2) then
         ctime(1:8) = ttime
         cdate(1:8) = tdate
         ndate      = 8
       endif

       if (iform .eq. 3) then
         ctime(1:8) = ttime
         if (tdate(1:2) .eq. '01') cdate(1:8)  = 'January '
         if (tdate(1:2) .eq. '02') cdate(1:9)  = 'February '
         if (tdate(1:2) .eq. '03') cdate(1:6)  = 'March '
         if (tdate(1:2) .eq. '04') cdate(1:6)  = 'April '
         if (tdate(1:2) .eq. '05') cdate(1:4)  = 'May '
         if (tdate(1:2) .eq. '06') cdate(1:5)  = 'June '
         if (tdate(1:2) .eq. '07') cdate(1:5)  = 'July '
         if (tdate(1:2) .eq. '08') cdate(1:7)  = 'August '
         if (tdate(1:2) .eq. '09') cdate(1:10) = 'September '
         if (tdate(1:2) .eq. '10') cdate(1:8)  = 'October '
         if (tdate(1:2) .eq. '11') cdate(1:9)  = 'November '
         if (tdate(1:2) .eq. '12') cdate(1:9)  = 'December '
         i = lastc ( cdate )
         cdate(i+2:i+3) = tdate(4:5)
         cdate(i+4:i+7) = ', 20'
         cdate(i+8:i+9) = tdate(7:8)
         ndate          = i + 9
       endif

       if (iform .eq. 4) then
         ctime(1:8)  = ttime
         cdate(1:8)  = tdate
         cdate(1:2)  = tdate(4:5)
         cdate(4:5)  = tdate(1:2)
         cdate(9:10) = tdate(7:8)
         cdate(7:8)  = '20'
         ndate       = 10
       endif

       ctime(3:3) = ':'
       ctime(6:6) = ':'
       if (cdate(1:1) .eq. ' ') cdate(1:1) = '0'
       if (cdate(4:4) .eq. ' ') cdate(4:4) = '0'

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       function timer 
       
     &          ( ttype, otime )
!__________________________________________________________________________________________________________________________________
!
!  PURPOSE: Timer for measuring execution time and CPU time. This code has been extracted from the main code of ZEUS-3D into this 
!  function in order to simplify timing of various parts of the code. 
!
!  INPUT VARIABLES:
!    ttype       'cpu' - to measure CPU time, 'wal' - to measure wall clock time
!    otime       previously taken measurement that will be subtracted from the new system timer value, giving time interval
!                (otime can be zero).
!
!  OUTPUT VARIABLES:
!    timer       time interval or initial timer value (for otime = 0.0d0).
!__________________________________________________________________________________________________________________________________
!
       character*3   ttype
       character*8   crdate
       character*10  crtime
       real          cptime
       real*8        timer, otime, ryear, rmon, rday, rhour, rmin, rsec, rtime
!__________________________________________________________________________________________________________________________________
!
       if (ttype .eq. 'cpu') then
         call cpu_time ( cptime )
         rtime = dble ( cptime )
       else
         call date_and_time ( crdate, crtime )
         read (crdate(3:4),*) ryear
         read (crdate(5:6),*) rmon
         read (crdate(7:8),*) rday
         read (crtime(1:2),*) rhour
         read (crtime(3:4),*) rmin
         read (crtime(5:10),*) rsec
         rtime = 31536000.0d0 * ryear + 2629822.96584d0 * rmon + 86400.0d0 * rday + 3600.0d0 * rhour + 60.0d0 * rmin + rsec
       endif
       timer = rtime - otime

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       function lastc 
       
     &          ( string )
!__________________________________________________________________________________________________________________________________
!
! Returns position of a last non-blank character in a string.
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       character*(*) string
       integer       lastc, ifinal, i, j
!__________________________________________________________________________________________________________________________________
!
       ifinal = len ( string )
       do i=1,ifinal
         j = ifinal + 1 - i
         if (string(j:j) .ne. ' ') goto 10
       enddo
       j = 0
10     continue
       lastc = j

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       function firstb 
       
     &          ( string )
!__________________________________________________________________________________________________________________________________
!
! Returns position of the first blank (" ") in a string.
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       character*(*) string
       integer       firstb, ifinal, i
!__________________________________________________________________________________________________________________________________
!
       ifinal = len ( string )
       do i=1,ifinal
         if (string(i:i) .eq. ' ') goto 10
       end do
       i = 0
10     continue
       firstb = i

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       function lastd 
       
     &          ( string )
!__________________________________________________________________________________________________________________________________
!
! Returns position of last dot (".") character in a string.
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       character*(*) string
       integer       lastd, ifinal, i, j
!__________________________________________________________________________________________________________________________________
!
       ifinal = len ( string )
       do i=1,ifinal
         j = ifinal + 1 - i
         if (string(j:j) .eq. '.') goto 10
       end do
       j = 0
10     continue
       lastd = j

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       function lasts ( string )
!__________________________________________________________________________________________________________________________________
!
! Returns position of last slash ("/") character in a string.
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       character*(*) string
       integer       lasts, ifinal, i, j
!__________________________________________________________________________________________________________________________________
!
       ifinal = len ( string )
       do i=1,ifinal
         j = ifinal + 1 - i
         if (string(j:j) .eq. '/') goto 10
       end do
       j = 0
10     continue
       lasts = j

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       function lastbs 
       
     &          ( string )
!__________________________________________________________________________________________________________________________________
!
! Returns position of last backslash ("\") character in a string.
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       character*(*) string
       integer       lastbs, ifinal, i, j
!__________________________________________________________________________________________________________________________________
!
       ifinal = len ( string )
       do i=1,ifinal
         j = ifinal + 1 - i
         if (string(j:j) .eq. '\') goto 10
       end do
       j = 0
10     continue
       lastbs = j

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       function lastp
       
     &          ( string )
!__________________________________________________________________________________________________________________________________
!
! Returns position of last plus ("+") character in a string.
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       character*(*) string
       integer       lastp, ifinal, i, j
!__________________________________________________________________________________________________________________________________
!
       ifinal = len ( string )
       do i=1,ifinal
         j = ifinal + 1 - i
         if (string(j:j) .eq. '+') goto 10
       end do
       j = 0
10     continue
       lastp = j

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       function last
       
     &          ( sub, string )
!__________________________________________________________________________________________________________________________________
!
! Returns position of the last instance of a substring in a string.
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       character*(*) sub, string
       integer       last, isub, ifinal, i, j, je
!__________________________________________________________________________________________________________________________________
!
       isub = len ( sub ) - 1
       ifinal = len ( string ) - isub
       do i=1,ifinal
         j = ifinal + 1 - i
         je = j + isub
         if (string(j:je) .eq. sub) goto 10
       enddo
       j = 0
10     continue
       last = j

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine sortarr 
       
     &            ( n, x, y )
!__________________________________________________________________________________________________________________________________
!
! Sorts array x(j) in ascending order and rearranges array y(j) according to x(j).
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       integer       i, j, n
       real*8        x(*), y(*), a, b
!__________________________________________________________________________________________________________________________________
!
       do j=2,n
         a = y(j)
         b = x(j)
         do i=j-1,1,-1
           if (x(i) .le. b) then
             goto 10
           else
             y(i+1) = y(i)
             x(i+1) = x(i)
           end if
         enddo
         i = 0
 10      continue
         y(i+1) = a
         x(i+1) = b
       enddo

       return
       end

!!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
!
!       function medapprox
!      
!     &          ( npts, xdata, leftend, rightend )
!!__________________________________________________________________________________________________________________________________
!!
!! Adapted by A. Men'shchikov from a code for the binapprox algorithm by Ryan J. Tibshirani (2008). 
!! Webpage: http://www.stat.cmu.edu/~ryantibs/median/ Paper: http://www.stat.cmu.edu/~ryantibs/papers/median.pdf
!!__________________________________________________________________________________________________________________________________
!!
!       implicit      none
!       integer       nbins
!       parameter   ( nbins = 1000 )
!       integer       i, j, nb, kmed, npts, leftcount, ntotcount, ncounts(0:nbins)
!       real*8        medapprox, scalefac, leftend, rightend, xdata(npts)
!!__________________________________________________________________________________________________________________________________
!!
!       if (leftend .eq. rightend) then
!         medapprox = leftend
!         return
!       endif
!       scalefac = dble ( nbins ) / (rightend - leftend)
!       leftcount = 0
!       do i=0,nbins
!         ncounts(i) = 0
!       enddo
!
!       do i=1,npts
!         if (xdata(i) .lt. leftend) then
!           leftcount = leftcount + 1
!         else if (xdata(i) .lt. rightend) then
!           nb = int ( (xdata(i) - leftend ) * scalefac )
!           ncounts(nb) = ncounts(nb) + 1
!         endif
!       enddo
!
!       if (mod ( npts, 2 ) .eq. 1) then
!       
!! Odd number 'npts' of data set elements, find the bin that contains the median.
!       
!         kmed = (npts + 1) / 2
!         ntotcount = leftcount
!         do i=0,nbins
!           ntotcount = ntotcount + ncounts(i)
!           if (ntotcount .ge. kmed) then
!             medapprox = dble ( 2 * i + 1 ) / (2.0d0 * scalefac) + leftend
!             exit
!           endif
!         enddo
!       else
!       
!! Even number 'npts' of data set elements, find the bins that contain the medians.
!       
!         kmed = npts / 2
!         ntotcount = leftcount
!         do i=0,nbins
!           ntotcount = ntotcount + ncounts(i)
!           if (ntotcount .ge. kmed) then
!             j = i
!   1         continue
!             if (ntotcount .eq. kmed) then
!               j = j + 1
!               ntotcount = ntotcount + ncounts(j)
!               goto 1
!             endif
!             medapprox = dble ( i + j + 1 ) / (2.0d0 * scalefac) + leftend
!             exit
!           endif
!         enddo
!       endif
!
!       return
!       end
!
!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
      
       function selectk

     &          ( k, n, arr )
!__________________________________________________________________________________________________________________________________
!
! From Numerical Recipes for FORTRAN. Recoded in a better language by A. Men'shchikov. Used to compute median value.
!
! Returns the k-th smallest value in the array arr(1:n). The input array will be rearranged to have this value in location arr(k),
! with all smaller elements moved to arr(1:k-1) (in arbitrary order) and all larger elements in arr(k+1..n) (in arbitrary order).
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       integer       k, n, i, ir, j, l, mid, l1, irl
       real*8        selectk, arr(n), a, temp
!__________________________________________________________________________________________________________________________________
!
       l = 1
       ir = n
 1     continue
       irl = ir - l
       if (irl .le. 1) then           ! Active partition contains 1 or 2 elements.
         if (irl .eq. 1) then         ! Active partition contains 2 elements.
           if (arr(ir) .lt. arr(l)) then
             temp = arr(l)
             arr(l) = arr(ir)
             arr(ir) = temp
           endif
         endif
         selectk = arr(k)
         goto 9
       else
         l1 = l + 1
         mid = (l + ir) / 2           ! Choose median of left, center, and right elements as partitioning element a.
         temp = arr(mid)              ! Also rearrange so that arr(l) <= arr(l+1), arr(ir) >= arr(l+1).
         arr(mid) = arr(l1)
         arr(l1) = temp
         if (arr(l) .gt. arr(ir)) then
           temp = arr(l)
           arr(l) = arr(ir)
           arr(ir) = temp
         endif
         if (arr(l1) .gt. arr(ir)) then
           temp = arr(l1)
           arr(l1) = arr(ir)
           arr(ir) = temp
         endif
         if (arr(l) .gt. arr(l1)) then
           temp = arr(l)
           arr(l) = arr(l1)
           arr(l1) = temp
         endif

! Initialize pointers for partitioning.

         a = arr(l1)                ! Partitioning element.
         i = l1
         j = ir
 3       continue
         i = i + 1
         do while (arr(i) .lt. a)   ! Scan up to find element > a.
           i = i + 1
         enddo
         j = j - 1
         do while (arr(j) .gt. a)   ! Scan down to find element < a.
           j = j - 1
         enddo
         if (j .ge. i) then         ! Pointers crossed. Exit with partitioning complete.
           temp = arr(i)  
           arr(i) = arr(j)          ! Exchange elements.
           arr(j) = temp
           goto 3
         endif

! Pointers crossed. Exit with partitioning complete.

         arr(l1) = arr(j)           ! Insert partitioning element.
         arr(j) = a
         if (j .ge. k) ir = j - 1   ! Keep active the partition that contains the k-th element.
         if (j .le. k) l = i
       endif
       goto 1

 9     return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine histogrm
       
     &            ( nf, nfun, ipmx, nfu, xn, fun, ibeg, iend, xtype )
!__________________________________________________________________________________________________________________________________
!
! Convert functions [nf..nfun] into histograms (for linear or logarithmic x-axis).
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       character*(*) xtype
       integer       m, i, l, k, nf, nfun, nfu, ipmx
       integer       ibeg(nfu), iend(nfu)
       real          fact, shft, xn(ipmx,nfu), fun(ipmx,nfu)
!__________________________________________________________________________________________________________________________________
!
       shft = 0.0
       fact = 1.0
       do m=nf,nfun
         if (2 * iend(m) + 1 .gt. ipmx) then
           return
         endif
         if (xtype .eq. 'lin') then
           shft = (xn(ibeg(m)+1,m) - xn(ibeg(m),m)) / 2.0
         else
           fact = sqrt ( xn(ibeg(m)+1,m) / xn(ibeg(m),m) )
         endif
         do i=iend(m),ibeg(m),-1
           l = 2 * i
           k = min ( i+1, iend(m) )
           xn (l  ,m) = xn (k,m)
           xn (l-1,m) = xn (i,m)
           fun(l  ,m) = fun(i,m)
           fun(l-1,m) = fun(i,m)
         enddo
         iend(m) = 2 * iend(m) + 1
         if (xtype .eq. 'lin') then
           xn(iend(m),m) = xn(iend(m)-2,m) + 2.0 * shft
         else
           xn(iend(m),m) = xn(iend(m)-2,m) * fact**2
         endif
         fun(iend(m),m) = fun(iend(m)-1,m)
         do l=ibeg(m),iend(m)
           if (xtype .eq. 'lin') then
             xn(l,m) = xn(l,m) - shft
           else
             xn(l,m) = xn(l,m) / fact
           endif
         enddo
       enddo

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine ps2pdf 
       
     &            ( psfile )
!__________________________________________________________________________________________________________________________________
!
! Converts PS file to PDF format by calling GhostScript's script (assuming Ghostscript installed and the ps2pdf script exists).
!__________________________________________________________________________________________________________________________________
!
!!!notforgnu       use ifport, only: system, ierrno
       use ifport, only: system, ierrno
       implicit      none
       character*(*) psfile
       integer       returncode, errnumber, ilc, lastc
       external      lastc
!__________________________________________________________________________________________________________________________________
!
       ilc = lastc ( psfile )
       returncode = system ( 'ps2pdf -sPAPERSIZE=a4 '//psfile//' '//psfile(1:ilc-3)//'.pdf' )
       if (returncode .eq. -1) then
         errnumber = ierrno ( )
         write (*,'(a,i3,a)') '   PS2PDF: Error converting '//psfile//' to PDF (error:', errnumber, ')...'
       else
         returncode = system ( '\rm '//psfile )
         if (returncode .eq. -1) then
           errnumber = ierrno ( )
           write (*,'(a,i3,a)') '   PS2PDF: Error deleting '//psfile//' (error:', errnumber, ')...'
         endif
       endif
       if (returncode .ne. 0) stop 999
         
       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine trimpdf 
       
     &            ( pdffile )
!__________________________________________________________________________________________________________________________________
!
! Trims PDF file of wite margins by calling GhostScript's script (assuming Ghostscript installed and the script exists).
!__________________________________________________________________________________________________________________________________
!
!!!notforgnu       use ifport, only: system, ierrno
       use ifport, only: system, ierrno
       implicit      none
       character*(*) pdffile
       integer       returncode, errnumber, ilc, lastc
       external      lastc
!__________________________________________________________________________________________________________________________________
!
       ilc = lastc ( pdffile )
       returncode = system ( 'trimpdf '//pdffile )
       if (returncode .eq. -1) then
         errnumber = ierrno ( )
         write (*,'(a,i3,a)') '   PS2PDF: Error trimming PDF file '//pdffile//' (error:', errnumber, ')...'
       else
         returncode = system ( '\rm '//pdffile )
         if (returncode .eq. -1) then
           errnumber = ierrno ( )
           write (*,'(a,i3,a)') '   PS2PDF: Error deleting '//pdffile//' (error:', errnumber, ')...'
         endif
       endif
         
       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine fitfun
       
     &            ( iotty, iolog, ma, npt, x, y, sig, a, ia, covar, alpha, chisq, iter, lgood, lconv, lsing, cmodel, modelfun
     &            , cwhat )
!__________________________________________________________________________________________________________________________________
!
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       logical       lsingular, lgood, lconv, lsing
       character*1   cma, crc, crcx
       character*2   cnpt
       character*3   citer
       character*6   lverbose
       character*9   chead(8)
       character*(*) cmodel, cwhat
       integer       i, ma, maa, ia(ma), iconverging, iter, npt, lastc, iotty, iolog, itermax, nit, nh, nhm3
       real*8        x(npt), y(npt), sig(npt), a(ma), covar(ma,ma), alpha(ma,ma), fit(1000), tau(1000), amin(10), amax(10), ao(10)
     &             , dfda(10), alamda, alamdao, chisq, ochisq, dchisq, xnu, freq0, wave0, kappa0, amuxmuH2, speedolight, dust2gas
     &             , da, damx, almostzero, pc
       parameter   ( amuxmuH2 = 1.6605402D-24 * 2.8d0, speedolight = 2.99792458d10, pc = 3.085678d18, almostzero = 1.0d-30
     &             , damx = 0.2d0, itermax = 99 )
       common / copacity / freq0, wave0, kappa0, dust2gas, lverbose
       external      lastc, modelfun
!__________________________________________________________________________________________________________________________________
!
       iter = 0
       chisq = 0.0d0
       dchisq = 0.0d0
       alamda = 0.0d0
       crc = ' '
       crcx = crc

! Compute a number of the degrees of freedom. For good results, CHI^2 should be approximately equal to XNU.

       maa = 0
       do i=1,ma
         maa = maa + ia(i)
       enddo
       xnu = npt - maa

! Set some reasonable limits on the parameter values.

       if (cmodel .eq. 'modbbody') then
         chead(1) = ' TEMPDUST'
         amin(1) = 3.0d0
         amax(1) = 2000.0d0 
         if (a(5) .gt. 1.0001d0 * pc) then
           chead(2) = ' TOTLMASS'
           amin(2) = 1.0d25
           amax(2) = 1.0d37
         else
           chead(2) = '  COLDENS'
           amin(2) = 1.0d15
           amax(2) = 1.0d30
         endif
         chead(3) = ' OPASLOPE'
         amin(3) = 0.1d0
         amax(3) = 1.0d1
         chead(4) = ' SLDANGLE'
         amin(4) = 1.0d-16
         amax(4) = 1.0d+01
         chead(5) = ' DISTANCE'
         amin(5) = 1.0d18
         amax(5) = 1.0d22
         chead(6) = '   CHISQ '
         chead(7) = '  DCHISQ '
         chead(8) = '  ALAMDA '
         nh = 8
         nhm3 = nh - 3
       endif
       if (cmodel .eq. 'thinbody') then
         chead(1) = ' TEMPDUST'
         amin(1) = 3.0d0
         amax(1) = 2000.0d0 
         if (a(4) .gt. 1.0001d0 * pc) then
           chead(2) = ' TOTLMASS'
           amin(2) = 1.0d25
           amax(2) = 1.0d37
         else
           chead(2) = '  COLDENS'
           amin(2) = 1.0d15
           amax(2) = 1.0d30
         endif
         chead(3) = ' OPASLOPE'
         amin(3) = 0.1d0
         amax(3) = 1.0d1
         chead(4) = ' DISTANCE'
         amin(4) = 1.0d18
         amax(4) = 1.0d22
         chead(5) = '   CHISQ '
         chead(6) = '  DCHISQ '
         chead(7) = '  ALAMDA '
         nh = 7
         nhm3 = nh - 3
       endif

       if (cwhat .eq. 'finalfit' .and. lverbose .ne. '-verb0') then
         if (iotty .gt. 0) write (iotty,'(/9a)') '   FITFUN:  IT ', (chead(i), i=1,nh)
         if (iolog .gt. 0) write (iolog,'(/9a)') '   FITFUN:  IT ', (chead(i), i=1,nh)
         if (iotty .gt. 0) write (iotty,'(/a,i4,a1,1p8e9.2)') '   FITFUN:', iter, crcx, (a(i), i=1,nhm3), chisq, dchisq, alamda
         if (iolog .gt. 0) write (iolog,'(/a,i4,a1,1p8e9.2)') '   FITFUN:', iter, crcx, (a(i), i=1,nhm3), chisq, dchisq, alamda
       endif

       do i=1,ma
         ao(i) = a(i)
       enddo
       alamda = -1.0d0
       lsingular = .false.

       call mrqmin ( x, y, sig, npt, a, ia, ma, covar, alpha, ma, chisq, modelfun, alamda, crc )

       if (crc .ne. ' ') then
         lsingular = .true.
         crcx = crc
       endif

       do i=1,ma
         da = (a(i) - ao(i)) / (ao(i) + 1.0d-30)
         if (abs ( da ) .gt. damx) a(i) = ao(i) * (1.0d0 + sign ( damx, da ))
       enddo

       do i=1,ma
         a(i) = min ( max ( a(i), amin(i) ), amax(i) )
         ao(i) = a(i)
       enddo

       iter = 1
       iconverging = 0

       if (cwhat .eq. 'finalfit' .and. lverbose .ne. '-verb0') then
         if (iotty .gt. 0) write (iotty,'(a,i4,a1,1p8e9.2)') '   FITFUN:', iter, crcx, (a(i), i=1,nhm3), chisq, dchisq, alamda
         if (iolog .gt. 0) write (iolog,'(a,i4,a1,1p8e9.2)') '   FITFUN:', iter, crcx, (a(i), i=1,nhm3), chisq, dchisq, alamda
       endif
       
 10    continue
         iter = iter + 1
         ochisq = chisq
         alamdao = alamda

         call mrqmin ( x, y, sig, npt, a, ia, ma, covar, alpha, ma, chisq, modelfun, alamda, crc )
     
         if (crc .ne. ' ') then
           lsingular = .true.
           crcx = crc
         endif

         dchisq = chisq - ochisq
         if (ochisq .gt. almostzero) then
           dchisq = dchisq / ochisq
         endif

         if (alamda .lt. alamdao .or. (dchisq .gt. -1.0d-3 .and. dchisq .lt. -almostzero)) then
           iconverging = iconverging + 1
         else
           if (abs ( dchisq ) .gt. almostzero) iconverging = 0
         endif

         do i=1,ma
           a(i) = min ( max ( a(i), amin(i) ), amax(i) )
           ao(i) = a(i)
         enddo
         
         if (cwhat .eq. 'finalfit' .and. lverbose .eq. '-verb2') then
           if (iotty .gt. 0) write (iotty,'(a,i4,a1,1p8e9.2)') '   FITFUN:', iter, crcx, (a(i), i=1,nhm3), chisq, dchisq, alamda
           if (iolog .gt. 0) write (iolog,'(a,i4,a1,1p8e9.2)') '   FITFUN:', iter, crcx, (a(i), i=1,nhm3), chisq, dchisq, alamda
         endif

         if (.not.lsingular .and. iter .lt. itermax .and. (alamda .lt. alamdao .or. alamda .lt. 1.0d3) .and. 
     &       (iconverging .lt. 4 .or. alamda .gt. 1.0d-5)) goto 10

         if (cwhat .eq. 'finalfit' .and. lverbose .eq. '-verb1') then
           if (iotty .gt. 0) write (iotty,'(a,i4,a1,1p8e9.2)') '   FITFUN:', iter, crcx, (a(i), i=1,nhm3), chisq, dchisq, alamda
           if (iolog .gt. 0) write (iolog,'(a,i4,a1,1p8e9.2)') '   FITFUN:', iter, crcx, (a(i), i=1,nhm3), chisq, dchisq, alamda
         endif
       continue

       lsing = lsingular
       if (cwhat .ne. 'finalfit') then
         if (lsing .or. (alamda .ge. alamdao .and. alamda .ge. 1.0d3)) iter = itermax
       endif
       lgood = .not.lsingular .and. chisq / (xnu + 1.0d0) .le. 1.0d0
       lconv = .not.lsingular .and. iter .lt. itermax

       alamda = 0.0d0

       call mrqmin ( x, y, sig, npt, a, ia, ma, covar, alpha, ma, chisq, modelfun, alamda, crc )

       if (cwhat .eq. 'finalfit' .and. lsingular) then
         if (iotty .gt. 0) write (iotty,'(/a)') '   FITFUN: ('//crcx//') Fitting aborted in GAUSSJ with a singular matrix.'
         if (iolog .gt. 0) write (iolog,'(/a)') '   FITFUN: ('//crcx//') Fitting aborted in GAUSSJ with a singular matrix.'
       endif

       if (cwhat .eq. 'finalfit' .and. lverbose .eq. '-verb2') then
         write (cma,'(i1)') nhm3
         if (iotty .gt. 0) write (iotty,'(/a,1p'//cma//'e9.2,a)') '   FITFUN: Abs±', (sqrt ( abs ( covar(i,i) ) ) 
     &                                                          , i=1,nhm3), ' <~ absolute uncertainties'
         if (iolog .gt. 0) write (iolog,'(/a,1p'//cma//'e9.2,a)') '   FITFUN: Abs±', (sqrt ( abs ( covar(i,i) ) ) 
     &                                                          , i=1,nhm3), ' <~ absolute uncertainties'
         if (iotty .gt. 0) write (iotty,'( a,1p'//cma//'e9.2,a)') '   FITFUN: Rel±', (sqrt ( abs ( covar(i,i) ) ) 
     &                                                          / (a(i) + 1.0d-30), i=1,nhm3), ' <~ relative uncertainties'
         if (iolog .gt. 0) write (iolog,'( a,1p'//cma//'e9.2,a)') '   FITFUN: Rel±', (sqrt ( abs ( covar(i,i) ) ) 
     &                                                          / (a(i) + 1.0d-30), i=1,nhm3), ' <~ relative uncertainties'
       endif

!  For good results, chi^2 should be approximately equal to xnu.

       if (cwhat .eq. 'finalfit' .and. lverbose .ne. '-verb0') then
         nit = 1
         write (citer,'(i3)') iter
         if (iter .gt. 0) nit = int ( log10 ( dble ( iter ) ) ) + 1
         if (chisq / (xnu + 1.0d0) .le. 1.0d0) then
           if (iotty .gt. 0) write (iotty,'(/a,2(1pe9.2,a),a)') '   FITFUN: chi^2:', chisq, ' xnu:', xnu, ' <~ GOOD FIT: '
     &                                                        //citer(3-nit+1:3)// ' iterations'
           if (iolog .gt. 0) write (iolog,'(/a,2(1pe9.2,a),a)') '   FITFUN: chi^2:', chisq, ' xnu:', xnu, ' <~ GOOD FIT: '
     &                                                        //citer(3-nit+1:3)// ' iterations'
         else
           if (iotty .gt. 0) write (iotty,'(/a,2(1pe9.2,a),a)') '   FITFUN: chi^2:', chisq, ' xnu:', xnu, ' <~ BAD FIT: '
     &                                                        //citer(3-nit+1:3)// ' iterations'
           if (iolog .gt. 0) write (iolog,'(/a,2(1pe9.2,a),a)') '   FITFUN: chi^2:', chisq, ' xnu:', xnu, ' <~ BAD FIT: '
     &                                                        //citer(3-nit+1:3)// ' iterations'
         endif
       endif

! Output modbbody values.

       do i=1,npt
         call modelfun ( x(i), a, fit(i), dfda, ma )
       enddo
       if (cmodel .eq. 'modbbody') then
         do i=1,npt
           if (a(5) .gt. 1.0001d0 * pc) then 
             tau(i) = a(2) * dust2gas * kappa0 * (x(i) / freq0)**a(3) / a(4) / a(5)**2
           else
             tau(i) = a(2) * dust2gas * amuxmuH2 * kappa0 * (x(i) / freq0)**a(3)
           endif
         enddo
       endif

       write (cnpt,'(i2)') npt

       if (cwhat .eq. 'finalfit' .and. lverbose .eq. '-verb2') then
         if (iotty .gt. 0) write (iotty,'(/a,'//cnpt//'(1pe8.1),a)') '   FITFUN:     Wavelengths:'
     &                   , (1.0d4 * speedolight / x(i), i=1,npt), ' um'
         if (iolog .gt. 0) write (iolog,'(/a,'//cnpt//'(1pe8.1),a)') '   FITFUN:     Wavelengths:'
     &                   , (1.0d4 * speedolight / x(i), i=1,npt), ' um'
         if (iotty .gt. 0) write (iotty,'( a,'//cnpt//'(1pe8.1),a)') '   FITFUN:     Frequencies:', (x(i), i=1,npt), ' Hz'
         if (iolog .gt. 0) write (iolog,'( a,'//cnpt//'(1pe8.1),a)') '   FITFUN:     Frequencies:', (x(i), i=1,npt), ' Hz'
         if (iotty .gt. 0) write (iotty,'( a,'//cnpt//'(1pe8.1),a)') '   FITFUN: Original fluxes:', (y(i), i=1,npt), ' Jy'
         if (iolog .gt. 0) write (iolog,'( a,'//cnpt//'(1pe8.1),a)') '   FITFUN: Original fluxes:', (y(i), i=1,npt), ' Jy'
         if (iotty .gt. 0) write (iotty,'( a,'//cnpt//'(1pe8.1),a)') '   FITFUN: Modbbody fluxes:', (fit(i), i=1,npt), ' Jy'
         if (iolog .gt. 0) write (iolog,'( a,'//cnpt//'(1pe8.1),a)') '   FITFUN: Modbbody fluxes:', (fit(i), i=1,npt), ' Jy'
         if (cmodel .eq. 'modbbody') then
           if (iotty .gt. 0) write (iotty,'( a,'//cnpt//'(1pe8.1),a)') '   FITFUN:  Optical depths:', (tau(i), i=1,npt)
           if (iolog .gt. 0) write (iolog,'( a,'//cnpt//'(1pe8.1),a)') '   FITFUN:  Optical depths:', (tau(i), i=1,npt)
         endif
       endif

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine modbbody 
       
     &            ( freqx, a, y, dyda, na )
!__________________________________________________________________________________________________________________________________
!
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       character*6   lverbose
       integer       na
       real*8        freqx, y, arg, k_B, hh, cc, bbody, taux, factor, freq0, wave0, kappa0, expargm1, exparg, dbbda1, dfada2, dfada3
     &             , dfada4, dfada5, freq, exptaux, hhokB, amuxmuH2, constant, dust2gas, a4bbody, pc, a(na), exptau, dyda(na)
       parameter   ( k_B = 1.380658d-16, hh = 6.626d-27, cc = 2.99792458d+10, amuxmuH2 = 1.6605402D-24 * 2.8d0, hhokB = hh / k_B
     &             , constant = 1.0d23 * 2.0d0 * hh / cc**2, pc = 3.085678d18 )
       common / copacity / freq0, wave0, kappa0, dust2gas, lverbose
!__________________________________________________________________________________________________________________________________
!
! Construct a model function (modbbody) used in fitting.
! a(1): temperature, a(2): mass or column density, a(3): beta, a(4): solid angle, a(5): distance.
! Normally, beta and distance are kept fixed.

       arg = min ( hhokB * freqx / min ( max ( a(1), 3.0d0 ), 2.0d3 ), 1.0d2 )
       exparg = exp ( arg )
       expargm1 = exparg - 1.0d0
       bbody = constant * freqx**3 / expargm1
       dbbda1 = bbody / expargm1 * exparg * arg / a(1)
       factor = 1.0d0
       dfada2 = 0.0d0
       dfada3 = 0.0d0
       dfada4 = 0.0d0
       dfada5 = 0.0d0
       freq = freqx / freq0
       if (a(5) .gt. 1.0001d0 * pc) then 
         taux = a(2) * dust2gas * kappa0 * freq**a(3) / a(4) / a(5)**2
         if (taux .le. 1.0d2) then
           exptau = exp ( -taux )
           factor = 1.0d0 - exptau
           exptaux = exptau * taux
           dfada2 = exptaux / a(2)
           dfada3 = exptaux * log ( freq )
           dfada4 = (1.0d0 - exptaux) / a(4)
           dfada5 = -2.0d0 * exptaux / a(5)
         endif
       else
         taux = a(2) * dust2gas * amuxmuH2 * kappa0 * freq**a(3)
         if (taux .le. 1.0d2) then
           exptau = exp ( -taux )
           factor = 1.0d0 - exptau
           exptaux = exptau * taux
           dfada2 = exptaux / a(2)
           dfada3 = exptaux * log ( freq )
         endif
       endif

       a4bbody = a(4) * bbody
       y = a4bbody * factor

! Compute partial derivatives of the model function with respect to its parameters.

       dyda(1) = a(4) * factor * dbbda1
       dyda(2) = a4bbody * dfada2
       dyda(3) = a4bbody * dfada3
       dyda(4) = a4bbody * dfada4
       dyda(5) = a4bbody * dfada5

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine thinbody 

     &            ( freqx, a, y, dyda, na )
!__________________________________________________________________________________________________________________________________
!
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       character*6   lverbose
       integer       na
       real*8        freqx, y, arg, k_B, hh, cc, bbody, taux, freq0, wave0, kappa0, expargm1, exparg, dbbda1, dtada2, dtada3, dtada4
     &             , freq, hhokB, dust2gas, amuxmuH2, constant, pc, a(na), dyda(na)
       parameter   ( k_B = 1.380658d-16, hh = 6.626d-27, cc = 2.99792458d+10, amuxmuH2 = 1.6605402D-24 * 2.8d0, pc = 3.085678d18
     &             , constant = 1.0d23 * 2.0d0 * hh / cc**2, hhokB = hh / k_B )
       common / copacity / freq0, wave0, kappa0, dust2gas, lverbose
!__________________________________________________________________________________________________________________________________
!
! Construct a model function (thinbody) used in fitting.
! a(1): temperature, a(2): mass or column density, a(3): beta, a(4): distance.
! Normally, a(3) and a(4) are kept fixed.

       arg = min ( hhokB * freqx / min ( max ( a(1), 3.0d0 ), 2.0d3 ), 1.0d2 )
       exparg = exp ( arg )
       expargm1 = exparg - 1.0d0
       bbody = constant * freqx**3 / expargm1
       dbbda1 = bbody / expargm1 * exparg * arg / a(1)
       freq = freqx / freq0
       if (a(4) .gt. 1.0001d0 * pc) then 
         taux = a(2) * dust2gas * kappa0 * freq**a(3) / a(4)**2
       else
         taux = a(2) * dust2gas * amuxmuH2 * kappa0 * freq**a(3)
       endif
       dtada2 = taux / a(2)
       dtada3 = taux * log ( freq )
       if (a(4) .gt. 1.0001d0 * pc) then 
         dtada4 = -2.0d0 * taux / a(4)
       else
         dtada4 = 0.0d0
       endif
       y = taux * bbody

! Compute partial derivatives of the model function with respect to its parameters.

       dyda(1) = taux * dbbda1
       dyda(2) = bbody * dtada2
       dyda(3) = bbody * dtada3
       dyda(4) = bbody * dtada4

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine graybody 

     &            ( freqx, a, y, dyda, na )
!__________________________________________________________________________________________________________________________________
!
! Kept here for compatibility with GETSOURCES v1.
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       character*6   lverbose
       integer       na
       real*8        freqx, y, arg, k_B, hh, cc, bbody, taux, factor, freq0, wave0, kappa0, expargm1, exparg, dbbda1, dfada3, dfada4
     &             , freq, exptaux, hhokB, amuxmuH2, constant, dust2gas, a2bbody, a(na), exptau, dyda(na)
       parameter   ( k_B = 1.380658d-16, hh = 6.626d-27, cc = 2.99792458d+10, amuxmuH2 = 1.6605402D-24 * 2.8d0, hhokB = hh / k_B
     &             , constant = 1.0d23 * 2.0d0 * hh / cc**2 )
       common / copacity / freq0, wave0, kappa0, dust2gas, lverbose
!__________________________________________________________________________________________________________________________________
!
! Construct a model function (graybody) used in fitting.
! a(1): temperature, a(2): solid angle, a(3): column density, a(4): beta.
! Normally, a(4) is kept fixed.

       arg = min ( hhokB * freqx / min ( max ( a(1), 3.0d0 ), 2.0d3 ), 1.0d2 )
       exparg = exp ( arg )
       expargm1 = exparg - 1.0d0
       bbody = constant * freqx**3 / expargm1
       dbbda1 = bbody / expargm1 * exparg * arg / a(1)
       factor = 1.0d0
       dfada3 = 0.0d0
       dfada4 = 0.0d0
       if (na .eq. 4) then 
         freq = freqx / freq0
         taux = a(3) * dust2gas * amuxmuH2 * kappa0 * freq**a(4)
         if (taux .le. 1.0d2) then
           exptau = exp ( -taux )
           factor = 1.0d0 - exptau
           exptaux = exptau * taux
           dfada3 = exptaux / a(3)
           dfada4 = exptaux * log ( freq )
         endif
       endif
       a2bbody = a(2) * bbody
       y = a2bbody * factor

! Compute partial derivatives of the model function with respect to its parameters.

       dyda(1) = a(2) * factor * dbbda1
       dyda(2) = bbody * factor
       if (na .eq. 4) then 
         dyda(3) = a2bbody * dfada3
         dyda(4) = a2bbody * dfada4
       endif

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine covsrt 
       
     &            ( covar, npc, ma, ia, mfit )
!__________________________________________________________________________________________________________________________________
!
! From Numerical Recipes for FORTRAN.
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       integer       ma, mfit, npc, ia(ma), i, j, k
       real*8        covar(npc,npc), swap
!__________________________________________________________________________________________________________________________________
!
       do i=mfit+1,ma
         do j=1,i
           covar(i,j) = 0.0d0
           covar(j,i) = 0.0d0
         enddo
       enddo
       k = mfit
       do j=ma,1,-1
         if (ia(j) .ne. 0) then
           do i=1,ma
             swap = covar(i,k)
             covar(i,k) = covar(i,j)
             covar(i,j) = swap
           enddo
           do i=1,ma
             swap = covar(k,i)
             covar(k,i) = covar(j,i)
             covar(j,i) = swap
           enddo
           k = k - 1
         endif
       enddo

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine gaussj 
       
     &            ( a, n, np, b, m, mp, crc )
!__________________________________________________________________________________________________________________________________
!
! From Numerical Recipes for FORTRAN.
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       character*1   crc
       integer       nmax
       parameter   ( nmax = 50 )
       integer       m, mp, n, np, i, icol, irow, j, k, l, ll, indxc(nmax), indxr(nmax), ipiv(nmax)
       real*8        big, dum, pivinv, a(np,np), b(np,mp)
!__________________________________________________________________________________________________________________________________
!
       crc = ' '
       do j=1,n
         ipiv(j) = 0
       enddo
       irow = 1
       icol = 1
       do i=1,n
         big = 0.0d0
         do j=1,n
           if (ipiv(j) .ne. 1) then
             do k=1,n
               if (ipiv(k) .eq. 0) then
                 if (abs ( a(j,k) ) .ge. big) then
                   big = abs ( a(j,k) )
                   irow = j
                   icol = k
                 endif
               elseif (ipiv(k) .gt. 1) then
                 crc = '$' !<-- Singular matrix: ipiv(k) > 1'
                 return
               endif
             enddo
           endif
         enddo
         ipiv(icol) = ipiv(icol) + 1
         if (irow .ne. icol) then
           do l=1,n
             dum = a(irow,l)
             a(irow,l) = a(icol,l)
             a(icol,l) = dum
           enddo
           do l=1,m
             dum = b(irow,l)
             b(irow,l) = b(icol,l)
             b(icol,l) = dum
           enddo
         endif
         indxr(i) = irow
         indxc(i) = icol
!!         if (a(icol,icol) .eq. 0.0d0) then
         if (abs ( a(icol,icol) ) .lt. 1.0d-99) then
           crc = '*' !<-- Singular matrix: a(icol,icol) = 0'
           return
         endif
         pivinv = 1.0d0 / a(icol,icol)
         a(icol,icol) = 1.0d0
         do l=1,n
           a(icol,l) = a(icol,l) * pivinv
         enddo
         do l=1,m
           b(icol,l) = b(icol,l) * pivinv
         enddo
         do ll=1,n
           if (ll .ne. icol) then
             dum = a(ll,icol)
             a(ll,icol) = 0.0d0
             do l=1,n
               a(ll,l) = a(ll,l) - a(icol,l) * dum
             enddo
             do l=1,m
               b(ll,l) = b(ll,l) - b(icol,l) * dum
             enddo
           endif
         enddo
       enddo
       do l=n,1,-1
         if (indxr(l) .ne. indxc(l)) then
           do k=1,n
             dum = a(k,indxr(l))
             a(k,indxr(l)) = a(k,indxc(l))
             a(k,indxc(l)) = dum
           enddo
         endif
       enddo

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine mrqcof 
       
     &            ( x, y, sig, ndata, a, ia, ma, alpha, beta, nalp, chisq, funcs )
!__________________________________________________________________________________________________________________________________
!
! From Numerical Recipes for FORTRAN.
!
! Used by MRQMIN to evaluate the linearized fitting matrix ALPHA and vector BETA, and calculate CHI^2.
!__________________________________________________________________________________________________________________________________
!
       implicit       none
       integer        mmax
       parameter    ( mmax = 5 )
       integer        ma, nalp, ndata, ia(ma), mfit, i, j, k, l, m
       real*8         chisq, a(ma), alpha(nalp,nalp), beta(ma), sig(ndata), x(ndata), y(ndata), dy, sig2i, wt, ymod, dyda(mmax)
       external       funcs
!__________________________________________________________________________________________________________________________________
!
       mfit = 0

       do j=1,ma
         if (ia(j) .ne. 0) mfit = mfit + 1
       enddo

! Initialize (symmetric) alpha, beta.

       do j=1,mfit
         do k=1,j
           alpha(j,k) = 0.0d0
         enddo
         beta(j) = 0.0d0
       enddo

! Summation loop over all data.

       chisq = 0.0d0

       do i=1,ndata

         call funcs ( x(i), a, ymod, dyda, ma )

         sig2i = 1.0d0 / (sig(i) * sig(i))

         dy = y(i) - ymod
         j = 0
         do l=1,ma
           if (ia(l) .ne. 0) then
             j = j + 1
             wt = dyda(l) * sig2i
             k = 0

!!         write (*,*) 'aaa', wt, dyda, sig2i

             do m=1,l
               if (ia(m) .ne. 0) then
                 k = k + 1
                 alpha(j,k) = alpha(j,k) + wt * dyda(m)
               endif
             enddo

!!         write (*,*) 'bbb'

             beta(j) = beta(j) + dy * wt
           endif
         enddo

! And find chi^2.

         chisq = chisq + dy * dy * sig2i
       enddo

! Fill in the symmetric side.

       do j=2,mfit
         do k=1,j-1
           alpha(k,j) = alpha(j,k)
         enddo
       enddo

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine mrqmin 
       
     &            ( x, y, sig, ndata, a, ia, ma, covar, alpha, nca, chisq, funcs, alamda, crc )
!__________________________________________________________________________________________________________________________________
!
! From Numerical Recipes for FORTRAN.
!
! Levenberg-Marquardt method, attempting to reduce the value chi^2 of a fit between a set of data points x(1:ndata), y(1:ndata)
! with individual standard deviations sig(1:ndata), and a non-linear function dependent on ma coefficients a(1:ma). The input
! array ia(1:ma) indicates by nonzero entries those components of 'a' that should be fitted for and by zero entries those
! components that should be held fixed at their input values. The program returns current best fit values for the parameters
! a(1:ma) and chi^2 = 'chisq'. The arrays covar(1:nca,1:nca), alpha(1:nca,1:nca) with physical dimension 'nca' (>= the number of
! fitted parameters) are used as working space during most iterations. Supply a subroutine funcs(x,a,yfit,dyda,ma) that evaluates
! the fitting function 'yfit' and its derivatives 'dyda' with respect to the fitting parameters 'a' at 'x'. On the first call
! provide an initial guess for the parameters 'a', and set 'alamda' < 0 for initialization (which then sets alamda=0.001). If a
! step succeeds 'chisq' becomes smaller and 'alamda' decreases by a factor of 10. You must call this routine repeatedly until
! convergence is achieved. Then, make one final call with alamda=0, so that covar(1:ma,1:ma) returns the covariance matrix, and
! 'alpha' the curvature matrix. (Parameters held fixed will return zero covariances).
!
! USES covsrt, gaussj, mrqcof
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       character*1   crc
       integer       mmax
       parameter   ( mmax = 20 )
       integer       ma, nca, ndata, ia(ma), j, k, l, mfit
       real*8        alamda, chisq, a(ma), alpha(nca,nca), covar(nca,nca), sig(ndata), x(ndata), y(ndata), ochisq
     &             , atry(mmax), beta(mmax), da(mmax)
       save          ochisq, atry, beta, da, mfit
       external      funcs
!__________________________________________________________________________________________________________________________________
!
! Initiaization.

       if (alamda .lt. 0.0d0) then
         mfit = 0
         do j=1,ma
           if (ia(j) .ne. 0) mfit = mfit + 1
         enddo
         alamda = 0.001d0

         call mrqcof ( x, y, sig, ndata, a, ia, ma, alpha, beta, nca, chisq, funcs )

         ochisq = chisq
         do j=1,ma
           atry(j) = a(j)
         enddo
       endif

! Alter linearized fitting matrix by augmenting diagonal elements.

       do j=1,mfit
         do k=1,mfit
           covar(j,k) = alpha(j,k)
         enddo
         covar(j,j) = alpha(j,j) * (1.0d0 + alamda)
         da(j) = beta(j)
       enddo

! Matrix solution.

       call gaussj ( covar, mfit, nca, da, 1, 1, crc )

! Once converged, evaluate covariance matrix.

       if (alamda .eq. 0.0d0) then

         call covsrt ( covar, nca, ma, ia, mfit )

! Spread out 'alpha' to its full size, too.

         call covsrt ( alpha, nca, ma, ia, mfit )

         return
       endif

! Did the trial succeed?

       j=0
       do l=1,ma
         if (ia(l) .ne. 0) then
           j = j + 1
           atry(l) = a(l) + da(j)
!!           atry(l) = max ( a(l) + da(j), 1.0d-90 )
         endif
       enddo

       call mrqcof ( x, y, sig, ndata, atry, ia, ma, covar, da, nca, chisq, funcs )

       if (chisq .lt. ochisq) then

! Success, accept the new solution.

         alamda = 0.1d0 * alamda
         ochisq = chisq
         do j=1,mfit
           do k=1,mfit
             alpha(j,k) = covar(j,k)
           enddo
           beta(j) = da(j)
         enddo
         do l=1,ma
           a(l) = atry(l)
         enddo
       else

! Failure, increase 'alamda' and return.

         alamda = 10.0d0 * alamda
         chisq = ochisq
       endif

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine spline
       
     &            ( x, y, n, yp1, ypn, y2 )
!__________________________________________________________________________________________________________________________________
!
! From Numerical Recipes for FORTRAN.
!
! Given arrays x(1:n) and y(1:n) containing a tabulated function, i.e., y_i = f(x_i), with x_1 < x_2 < ... < x_n, and given values
! yp1 and ypn for the first derivative of the interpolating function at points 1 and n, respectively, this routine returns an array
! y2(1:n) of length n which contains the second derivatives of the interpolating function at the tabulated points x_i. If yp1 and/or
! ypn are equal to 10^30 or larger, the routine is signaled to set the corresponding boundary condition for a natural spline, with
! with zero second derivative at that boundary. Parameter: nmax is the largest anticipated value of n.
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       integer       n, nmax, i, k
       parameter     ( nmax = 500 )
       real*8        yp1, ypn, x(n), y(n), y2(n), p, qn, sig, un, u(nmax)
!__________________________________________________________________________________________________________________________________
!
       if (yp1 .gt. 0.99d30) then
         y2(1) = 0.0d0
         u (1) = 0.0d0
       else
         y2(1) = -0.5d0
         u (1) = (3.0d0 / (x(2) - x(1))) * ((y(2) - y(1)) / (x(2) - x(1)) - yp1)
       endif
       do 11 i=2,n-1
         sig = (x(i) - x(i-1)) / (x(i+1) - x(i-1))
         p = sig * y2(i-1) + 2.0d0
         y2(i) = (sig - 1.0d0) / p
         u (i) = (6.0d0 * ((y(i+1) - y(i)) / (x(i+1) - x(i)) - (y(i) - y(i-1)) / (x(i) - x(i-1))) 
     &         / (x(i+1) - x(i-1)) - sig * u(i-1)) / p
 11    continue
       if (ypn .gt. 0.99d30) then
         qn = 0.0d0
         un = 0.0d0
       else
         qn = 0.5d0
         un = (3.0d0 / (x(n) - x(n-1))) * (ypn - (y(n) - y(n-1)) / (x(n) - x(n-1)))
       endif
       y2(n) = (un - qn * u(n-1)) / (qn * y2(n-1) + 1.0d0)
       do 12 k=n-1,1,-1
         y2(k) = y2(k) * y2(k+1) + u(k)
 12    continue

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine splint
       
     &            ( xa, ya, y2a, n, x, y )
!__________________________________________________________________________________________________________________________________
!
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       integer       n, k, khi, klo
       real*8        x, y, xa(n), y2a(n), ya(n), a, b, h
!__________________________________________________________________________________________________________________________________
!
       klo = 1
       khi = n

 1     if (khi - klo .gt. 1) then
         k = (khi + klo) / 2
         if (xa(k) .gt. x) then
           khi = k
         else
           klo = k
         endif
       goto 1
       endif
       h = xa(khi) - xa(klo)
       if (h .eq. 0.0d0) then
         write (*,*) 'bad xa input in splint'
         read (*,*)
       endif

       a = (xa(khi) - x) / h
       b = (x - xa(klo)) / h
       y = a * ya(klo) + b * ya(khi) + ((a**3 - a) * y2a(klo) + (b**3 - b) * y2a(khi)) * (h**2) / 6.0d0
      
       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       function ran1 
       
     &          ( idum )
!__________________________________________________________________________________________________________________________________
!
!    whp:zeus3d.ran1 <------------------- returns uniform random deviate
!    from numerical recipes                               december, 1992
!
!    written by: Press etal.
!    modified 1:
!
!  PURPOSE:  Returns a uniformly distrbuted random deviate between 0.0 and 1.0 (see Numerical Recipes, 1st edition for FORTRAN,
!            page 196).
!
!  INPUT VARIABLES:
!    idum        .ge. 0 => get next random number from sequence
!                .lt. 0 => initialise sequence depending on absolute value of "idum", and gets next random number
!
!  OUTPUT VARIABLES:
!    ran1        deviate with uniform distribution between 0.0 and 1.0
!
!  LOCAL VARIABLES:
!
!  EXTERNALS: [NONE]
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       integer       imax, m1, ia1, ic1, m2, ia2, ic2, m3, ia3, ic3
       real*8        rm1, rm2
       parameter     ( imax = 97 )
       parameter     ( m1 = 259200, ia1 = 7141, ic1 = 54733, rm1 = 1.0d0 / m1 )
       parameter     ( m2 = 134456, ia2 = 8121, ic2 = 28411, rm2 = 1.0d0 / m2 )
       parameter     ( m3 = 243000, ia3 = 4561, ic3 = 51349 )
       integer       idum, iff, ix1, ix2, ix3, i
       real*8        ran1
       real*8        r(imax)
       data          iff / 0 /
       save          iff, ix1, ix2, ix3, r
!__________________________________________________________________________________________________________________________________
!
!  If sequence is to be initialised, start here ...

       if (idum .lt. 0 .or. iff .eq. 0) then
         iff = 1
         ix1 = mod ( ic1 - idum     , m1 )
         ix1 = mod ( ia1 * ix1 + ic1, m1 )
         ix2 = mod ( ix1            , m2 )
         ix1 = mod ( ia1 * ix1 + ic1, m1 )
         ix3 = mod ( ix1            , m3 )
         do i=1,imax
           ix1  = mod ( ia1 * ix1 + ic1, m1 )
           ix2  = mod ( ia2 * ix2 + ic2, m2 )
           r(i) = ( dfloat ( ix1 ) + dfloat ( ix2 ) * rm2 ) * rm1
         enddo
         idum = 1
       endif

!  Otherwise, start here ....

       ix1  = mod ( ia1 * ix1 + ic1, m1 )
       ix2  = mod ( ia2 * ix2 + ic2, m2 )
       ix3  = mod ( ia3 * ix3 + ic3, m3 )
       i    = 1 + ( imax * ix3 ) / m3
       ran1 = r(i)
       r(i) = ( dfloat ( ix1 ) + dfloat ( ix2 ) * rm2 ) * rm1

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

      function gasdev
               
     &         ( idum )
!__________________________________________________________________________________________________________________________________
!
!__________________________________________________________________________________________________________________________________
!
      implicit none
      integer idum
      real*8 gasdev
cu    uses ran1
      integer iset
      real*8 fac,gset,rsq,v1,v2,ran1
      save iset,gset
      data iset /0/
!__________________________________________________________________________________________________________________________________
!
      if (iset .eq. 0) then
 1      continue
        v1 = 2.0d0 * ran1 ( idum ) - 1.0d0
        v2 = 2.0d0 * ran1 ( idum ) - 1.0d0
        rsq = v1**2 + v2**2
        if(rsq .ge. 1.0d0 .or. rsq .eq. 0.0d0) goto 1
        fac = sqrt ( -2.0d0 * log ( rsq ) / rsq )
        gset = v1 * fac
        gasdev = v2 * fac
        iset = 1
      else
        gasdev = gset
        iset = 0
      endif
      
      return
      end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine planck 
       
     &            ( t, fq, bf )
!__________________________________________________________________________________________________________________________________
!
! Planck function bf at temperature t and wave number fq = 1 / lambda.
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       real*8        t, fq, bf, x1, x2, expmax, xx, ex, c2, exmx
       data          x1 / 3.972968d-16 /, x2 / 1.438833d+00 /, expmax / 1.5d+02 /
!__________________________________________________________________________________________________________________________________
!
! here: x1=2*h*c, x2=h*c/k; 1/(exp{x}-1) => exp{-x}/(1-exp{-x})

       xx = x2 * fq / t

       if (xx .le. expmax) then
         ex = exp ( -xx )
         bf = x1 * fq**3 * ex / ( 1.0d0 - ex )
       else
         c2 = expmax / x2
         exmx = exp ( -expmax )
         bf = x1 * (c2 * t)**3 * exmx / ( 1.0d0 - exmx )
       endif

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine mess 
       
     &            ( message, crlf )
!__________________________________________________________________________________________________________________________________
!
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       integer       iotty, iolog
       character*(*) message, crlf
       common / lmess / iotty, iolog
!__________________________________________________________________________________________________________________________________
!
       if (iotty .gt. 0) then
         if (crlf .eq. 'a' .or. crlf .eq. 'ab') write (iotty,'(a)') '         :'
         write (iotty, '(a)') message
         if (crlf .eq. 'b' .or. crlf .eq. 'ab') write (iotty,'(a)') '         :'
       end if
       if (iolog .gt. 0) then
         if (crlf .eq. 'a' .or. crlf .eq. 'ab') write (iolog,'(a)') '         :'
         write (iolog, '(a)') message
         if (crlf .eq. 'b' .or. crlf .eq. 'ab') write (iolog,'(a)') '         :'
       end if

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine fillarea 
     
     &            ( lpartition, iseedx, iseedy, nextrmax, nextr, newpixval, method, zero, boundpixval, nx, ny, n1x, n2x, n1y, n2y
     &            , nxmin, nxmax, nymin, nymax, iworker, segmn, segmo, mxco, myco, deltx, delty, ntouching, nsx )
!__________________________________________________________________________________________________________________________________
!
! Alexander Men'shchikov, DAp IRFU CEA Saclay, France, on March 12, 2009
! Based on the Tint Fill Algorithm by Alvy Ray Smith, Computer Graphics Lab, New York Institute of Technology, Old Westbury,
! NY 11568. Published in: SIGGRAPH 79 Conference Proceedings, Aug 1979, 276-282. Also Technical Memo No 6, Computer Graphics Lab,
! New York Institute of Technology, Jul 1978, and issued as tutorial notes at SIGGRAPHs 78, 80-82.
!__________________________________________________________________________________________________________________________________
!                     
       implicit      none
       logical       logic, lpartition, lnxmin, lnxmax, lnymin, lnymax, lnochange
       character*(*) method
       integer       iseedx, iseedy, ix, iy, ilx, irx, maxstack, nx, ny, nstack, iyref, ilxref, irxref, nextrmax, nextr, n1y, n2y, n
     &             , nxmin, nxmax, nymin, nymax, iter, i, j, nxmino, nxmaxo, nymino, nymaxo
       parameter   ( maxstack = 10000000 )
       integer       ixstack(maxstack), iystack(maxstack), n1x(*), n2x(*), ntouching(nextrmax,*), nsx(*)
       real*8        newpixval, oldpixval, boundpixval, zero
       real*8        iworker(nx,ny), segmn(nx,ny), segmo(nx,ny), mxco(*), myco(*), deltx(*), delty(*)
       external      fillleft, fillright, scanhi, scanlo, pop, push
!__________________________________________________________________________________________________________________________________
!
       nstack = 0
       ix = iseedx
       iy = iseedy
       nxmin = ix
       nxmax = ix
       nymin = iy
       nymax = iy
       ilx = ix
       irx = ix

       oldpixval = iworker(ix,iy)

       n = nint ( newpixval )
       
       if (method .eq. 'positive' .or. method .eq. 'boundary' .or. 
     &     abs ( newpixval - oldpixval ) .gt. zero * abs ( newpixval + oldpixval )) then
         iyref = iy
         ilxref = ix
         irxref = ix

         call push ( ix, iy, ixstack, iystack, nstack, maxstack )
         
         do while (nstack .gt. 0)
       
           call pop ( ixstack, iystack, nstack, maxstack, ix, iy )
                                               
           if (abs ( newpixval - iworker(ix,iy) ) .gt. zero * abs ( newpixval + iworker(ix,iy) )) then
           
             call fillright ( lpartition, ix, iy, irx, nextrmax, nextr, oldpixval, newpixval, method, zero, boundpixval, nx, ny, n2x
     &                      , iworker, segmn, segmo, mxco, myco, deltx, delty, ntouching, nsx )
             
             call fillleft  ( lpartition, ix, iy, ilx, nextrmax, nextr, oldpixval, newpixval, method, zero, boundpixval, nx, ny, n1x
     &                      , iworker, segmn, segmo, mxco, myco, deltx, delty, ntouching, nsx )
             
             logic = ilx .ge. ilxref - 1 .and. irx .le. irxref + 1

             if (iy .eq. iyref + 1 .and. logic) then
                    
               call scanhi ( lpartition, iy, ilx, irx, nextrmax, nextr, oldpixval, newpixval, boundpixval, method, zero, ixstack
     &                     , iystack, nstack, maxstack, nx, ny, n2y, iworker, segmn, segmo, mxco, myco, deltx, delty, ntouching
     &                     , nsx )
             
             else if (iy .eq. iyref - 1 .and. logic) then
               
               call scanlo ( lpartition, iy, ilx, irx, nextrmax, nextr, oldpixval, newpixval, boundpixval, method, zero, ixstack
     &                     , iystack, nstack, maxstack, nx, ny, n1y, iworker, segmn, segmo, mxco, myco, deltx, delty, ntouching
     &                     , nsx )
             else
             
               call scanhi ( lpartition, iy, ilx, irx, nextrmax, nextr, oldpixval, newpixval, boundpixval, method, zero, ixstack
     &                     , iystack, nstack, maxstack, nx, ny, n2y, iworker, segmn, segmo, mxco, myco, deltx, delty, ntouching
     &                     , nsx )

               call scanlo ( lpartition, iy, ilx, irx, nextrmax, nextr, oldpixval, newpixval, boundpixval, method, zero, ixstack
     &                     , iystack, nstack, maxstack, nx, ny, n1y, iworker, segmn, segmo, mxco, myco, deltx, delty, ntouching
     &                     , nsx )
             endif
             
             iyref = iy
             ilxref = ilx
             irxref = irx
           endif
         enddo
       endif

! Determine the bounding box for the segmentation mask of a source.

       nxmin = min ( nxmin, ilx )
       nxmax = max ( nxmax, irx )
       nymin = min ( nymin, iy )
       nymax = max ( nymax, iy )

       lnxmin = .true.
       lnxmax = .true.
       lnymin = .true.
       lnymax = .true.

       do iter=1,nx*ny
       
! Expand vertical edges to find the box limits.

         if (lnxmin .or. lnymin .or. lnymax) then
           lnxmin = .false.
           nxmino = nxmin
           do j=nymin,nymax
             if (nint ( segmn(nxmin,j) ) .eq. n) then
               nxmin = max ( nxmin - 1, n1x(j) )
               if (nxmin .ne. nxmino) lnxmin = .true.
               exit
             endif
           enddo
         endif
         
         if (lnxmax .or. lnymin .or. lnymax) then
           lnxmax = .false.
           nxmaxo = nxmax
           do j=nymin,nymax
             if (nint ( segmn(nxmax,j) ) .eq. n) then
               nxmax = min ( nxmax + 1, n2x(j) )
               if (nxmax .ne. nxmaxo) lnxmax = .true.
               exit
             endif
           enddo
         endif
              
! Expand horizontal edges to find the box limits.

         if (lnymin .or. lnxmin .or. lnxmax) then
           lnymin = .false.
           nymino = nymin
           do i=nxmin,nxmax
             if (nint ( segmn(i,nymin) ) .eq. n) then
               nymin = max ( nymin - 1, n1y )
               if (nymin .ne. nymino) lnymin = .true.
               exit
             endif
           enddo
         endif
         
         if (lnymax .or. lnxmin .or. lnxmax) then
           lnymax = .false.
           nymaxo = nymax
           do i=nxmin,nxmax
             if (nint ( segmn(i,nymax) ) .eq. n) then
               nymax = min ( nymax + 1, n2y )
               if (nymax .ne. nymaxo) lnymax = .true.
               exit
             endif
           enddo
         endif

         lnochange = .not.lnxmin .and. .not.lnxmax .and. .not.lnymin .and. .not.lnymax
         if (lnochange) exit
       enddo
       
       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine fillright 
       
     &            ( lpartition, ix, iy, irx, nextrmax, nextr, oldpixval, newpixval, method, zero, boundpixval, nx, ny, n2x, iworker
     &            , segmn, segmo, mxco, myco, deltx, delty, ntouching, nsx )
!__________________________________________________________________________________________________________________________________
!
! Alexander Men'shchikov, DAp IRFU CEA Saclay, France, on March 12, 2009
! Based on the Tint Fill Algorithm by Alvy Ray Smith, Computer Graphics Lab, New York Institute of Technology, Old Westbury,
! NY 11568. Published in: SIGGRAPH 79 Conference Proceedings, Aug 1979, 276-282. Also Technical Memo No 6, Computer Graphics Lab,
! New York Institute of Technology, Jul 1978, and issued as tutorial notes at SIGGRAPHs 78, 80-82.
!__________________________________________________________________________________________________________________________________
!                     
       implicit      none
       logical       lpartition, lpixval
       character*(*) method
       integer       ix, iy, irx, nx, ny, mx, nextrmax, nextr, n2x(*), ntouching(nextrmax,*), nsx(*)
       real*8        oldpixval, newpixval, boundpixval, zero
       real*8        iworker(nx,ny), segmn(nx,ny), segmo(nx,ny), mxco(*), myco(*), deltx(*), delty(*)
!__________________________________________________________________________________________________________________________________
!
       mx = ix
       do while (mx .le. n2x(iy))
         if (lpixval ( lpartition, '.eq.', mx, iy, nextrmax, nextr, oldpixval, newpixval, boundpixval, method, zero, nx, ny, iworker
     &               , segmn, segmo, mxco, myco, deltx, delty, ntouching, nsx )) then 
           iworker(mx,iy) = newpixval
           segmn  (mx,iy) = newpixval
           mx = mx + 1
         else
           exit
         endif
       enddo
       irx = mx - 1
       
       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine fillleft 
       
     &            ( lpartition, ix, iy, ilx, nextrmax, nextr, oldpixval, newpixval, method, zero, boundpixval, nx, ny, n1x, iworker
     &            , segmn, segmo, mxco, myco, deltx, delty, ntouching, nsx )
!__________________________________________________________________________________________________________________________________
!
! Alexander Men'shchikov, DAp IRFU CEA Saclay, France, on March 12, 2009
! Based on the Tint Fill Algorithm by Alvy Ray Smith, Computer Graphics Lab, New York Institute of Technology, Old Westbury,
! NY 11568. Published in: SIGGRAPH 79 Conference Proceedings, Aug 1979, 276-282. Also Technical Memo No 6, Computer Graphics Lab,
! New York Institute of Technology, Jul 1978, and issued as tutorial notes at SIGGRAPHs 78, 80-82.
!__________________________________________________________________________________________________________________________________
!                     
       implicit      none
       logical       lpartition, lpixval
       character*(*) method
       integer       ix, iy, ilx, nx, ny, mx, nextrmax, nextr, n1x(*), ntouching(nextrmax,*), nsx(*)
       real*8        oldpixval, newpixval, boundpixval, zero
       real*8        iworker(nx,ny), segmn(nx,ny), segmo(nx,ny), mxco(*), myco(*), deltx(*), delty(*)
!__________________________________________________________________________________________________________________________________
!
       mx = ix - 1
       if (method .eq. 'boundary') mx = ix
       do while (mx .ge. n1x(iy))
         if (lpixval ( lpartition, '.eq.', mx, iy, nextrmax, nextr, oldpixval, newpixval, boundpixval, method, zero, nx, ny, iworker
     &               , segmn, segmo, mxco, myco, deltx, delty, ntouching, nsx )) then
           if (method .eq. 'boundary' .and. mx .eq. ix) mx = ix - 1
           iworker(mx,iy) = newpixval
           segmn  (mx,iy) = newpixval
           mx = mx - 1
         else
           exit
         endif
       enddo
       ilx = mx + 1
       
       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine scanhi 
       
     &            ( lpartition, iy, ilx, irx, nextrmax, nextr, oldpixval, newpixval, boundpixval, method, zero, ixstack, iystack
     &            , nstack, maxstack, nx, ny, n2y, iworker, segmn, segmo, mxco, myco, deltx, delty, ntouching, nsx )
!__________________________________________________________________________________________________________________________________
!
! Alexander Men'shchikov, DAp IRFU CEA Saclay, France, on March 12, 2009
! Based on the Tint Fill Algorithm by Alvy Ray Smith, Computer Graphics Lab, New York Institute of Technology, Old Westbury,
! NY 11568. Published in: SIGGRAPH 79 Conference Proceedings, Aug 1979, 276-282. Also Technical Memo No 6, Computer Graphics Lab,
! New York Institute of Technology, Jul 1978, and issued as tutorial notes at SIGGRAPHs 78, 80-82.
!__________________________________________________________________________________________________________________________________
!                     
       implicit      none
       logical       lpartition, lpixval
       character*(*) method
       integer       iy, nstack, nx, ny, ilx, irx, mx, my, maxstack, nextrmax, nextr, n2y
       integer       ixstack(maxstack), iystack(maxstack), ntouching(nextrmax,*), nsx(*)
       real*8        oldpixval, newpixval, boundpixval, zero
       real*8        iworker(nx,ny), segmn(nx,ny), segmo(nx,ny), mxco(*), myco(*), deltx(*), delty(*)
!__________________________________________________________________________________________________________________________________
!
       mx = ilx
       my = iy + 1
       if (my .le. n2y) then
         do while (mx .le. irx)
           do while (mx .le. irx)
             if (lpixval ( lpartition, '.ne.', mx, my, nextrmax, nextr, oldpixval, newpixval, boundpixval, method, zero, nx, ny
     &                   , iworker, segmn, segmo, mxco, myco, deltx, delty, ntouching, nsx )) then
               mx = mx + 1
             else
               exit
             endif
           enddo
           if (mx .gt. irx) exit

           call push ( mx, my, ixstack, iystack, nstack, maxstack )
           
           do while (mx .le. irx)
             if (lpixval ( lpartition, '.eq.', mx, my, nextrmax, nextr, oldpixval, newpixval, boundpixval, method, zero, nx, ny
     &                   , iworker, segmn, segmo, mxco, myco, deltx, delty, ntouching, nsx )) then
               mx = mx + 1
             else
               exit
             endif
           enddo
         enddo
       endif
       
       return
       end 
 
!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine scanlo 
       
     &            ( lpartition, iy, ilx, irx, nextrmax, nextr, oldpixval, newpixval, boundpixval, method, zero, ixstack, iystack
     &            , nstack, maxstack, nx, ny, n1y, iworker, segmn, segmo, mxco, myco, deltx, delty, ntouching, nsx )
!__________________________________________________________________________________________________________________________________
!
! Alexander Men'shchikov, DAp IRFU CEA Saclay, France, on March 12, 2009
! Based on the Tint Fill Algorithm by Alvy Ray Smith, Computer Graphics Lab, New York Institute of Technology, Old Westbury,
! NY 11568. Published in: SIGGRAPH 79 Conference Proceedings, Aug 1979, 276-282. Also Technical Memo No 6, Computer Graphics Lab,
! New York Institute of Technology, Jul 1978, and issued as tutorial notes at SIGGRAPHs 78, 80-82.
!__________________________________________________________________________________________________________________________________
!                     
       implicit      none
       logical       lpartition, lpixval
       character*(*) method
       integer       iy, nstack, nx, ny, ilx, irx, mx, my, maxstack, nextrmax, nextr, n1y
       integer       ixstack(maxstack), iystack(maxstack), ntouching(nextrmax,*), nsx(*)
       real*8        oldpixval, newpixval, boundpixval, zero
       real*8        iworker(nx,ny), segmn(nx,ny), segmo(nx,ny), mxco(*), myco(*), deltx(*), delty(*)
!__________________________________________________________________________________________________________________________________
!
       mx = ilx
       my = iy - 1
       if (my .ge. n1y) then
         do while (mx .le. irx)
           do while (mx .le. irx)
             if (lpixval ( lpartition, '.ne.', mx, my, nextrmax, nextr, oldpixval, newpixval, boundpixval, method, zero, nx, ny
     &                   , iworker, segmn, segmo, mxco, myco, deltx, delty, ntouching, nsx )) then
               mx = mx + 1
             else
               exit
             endif
           enddo
           if (mx .gt. irx) exit
         
           call push ( mx, my, ixstack, iystack, nstack, maxstack )
         
           do while (mx .le. irx)
             if (lpixval ( lpartition, '.eq.', mx, my, nextrmax, nextr, oldpixval, newpixval, boundpixval, method, zero, nx, ny
     &                   , iworker, segmn, segmo, mxco, myco, deltx, delty, ntouching, nsx )) then
               mx = mx + 1
             else
               exit
             endif
           enddo
         enddo
       endif
       
       return
       end 

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       function lpixval 
       
     &          ( lpartition, clogic, ix, iy, nextrmax, nextr, oldpixval, newpixval, boundpixval, method, zero, nx, ny, iworker
     &          , segmn, segmo, mxco, myco, deltx, delty, ntouching, nsx )
!__________________________________________________________________________________________________________________________________
!
! Alexander Men'shchikov, DAp IRFU CEA Saclay, France, on March 12, 2009
!__________________________________________________________________________________________________________________________________
!                     
       implicit      none
       logical       lpartition, lpixval, lntakenbyanother
       character*(*) method, clogic
       integer       ix, iy, nx, ny, nextrmax, newpixvali, n, k, l, mindist, nextr
       integer       ntouching(nextrmax,*), nsx(*)
       real*8        oldpixval, newpixval, boundpixval, zero, rxykn2min, rxyk2, dblix, dbliy
       real*8        iworker(nx,ny), segmn(nx,ny), segmo(nx,ny), mxco(*), myco(*), deltx(*), delty(*)
!__________________________________________________________________________________________________________________________________
!
       newpixvali = nint ( newpixval )                         

       k = newpixvali

       n = nint ( segmn(ix,iy) )

! If the segmentation pixel >= 1, this means that it has been assigned to a source
! and if the source number differs from the current filling number, then it's a different source: it's taken.

       if (segmn(ix,iy) .ge. 1.0d0 .and. k .ne. n) then
         lntakenbyanother = .true.
       else
         lntakenbyanother = .false.    !<-- n = 0 or newpixvali = n
       endif
              
! The next block works for existing touching sources whose pixels need to be partitioned.

       if (lpartition) then
         
         if (segmn(ix,iy) .ge. 1.0d0 .and. lntakenbyanother) then
                                                           
! If a source n is touching the source k, then consider the distances of the current pixel (ix,iy) from both sources
! to decide, whether the pixel should belong to k or to n.
                                                           
           if (nint ( iworker(ix,iy) ) .eq. n) then
             if (segmo(ix,iy) .lt. 1.0d0) then
               mindist = n
               do l=1,nextr
                 if (l .ne. n .and. ntouching(l,n) .eq. 1) then
                   dblix = dble ( ix )                           
                   dbliy = dble ( iy )                           
                   rxykn2min = (dblix - (mxco(n) + deltx(l)))**2 + (dbliy - (myco(n) + delty(l)))**2
                   rxyk2 = (dblix - mxco(l))**2 + (dbliy - myco(l))**2
                   if (rxyk2 .lt. rxykn2min) then                
                     mindist = l
                   endif
                 endif
               enddo
               if (mindist .eq. k) lntakenbyanother = .false.
             endif
           endif                                           
                                                           
! If a source n is merged with a more prominent source k, then its segmentation pixels are free to be used by the other.
                                                           
           if (nsx(n) .eq. 0) then
             if (ntouching(k,n) .eq. 2) then
               lntakenbyanother = .false.
             endif
           endif
         endif
       endif
       
       if (clogic .eq. '.ne.') then
              
! Pixel value is not good for taking the pixel in segmentation mask              
         
         if (method .eq. 'interior') then
           lpixval = abs ( iworker(ix,iy) - oldpixval   ) .gt. zero * abs ( iworker(ix,iy) + oldpixval )
         else if (method .eq. 'boundary') then
           lpixval = abs ( iworker(ix,iy) - boundpixval ) .le. zero * abs ( iworker(ix,iy) + oldpixval )
         else if (method .eq. 'positive') then
           lpixval = iworker(ix,iy) .le. zero .or. lntakenbyanother
         else if (method .eq. 'negative') then
           lpixval = iworker(ix,iy) .gt. zero
         else
           write (*,'(/a)') '   LPIXVAL: Invalid method: '//method
           stop 99
         endif

       else if (clogic .eq. '.eq.') then

! Pixel value is good for taking the pixel into segmentation mask              
         
         if (method .eq. 'interior') then
           lpixval = abs ( iworker(ix,iy) - oldpixval   ) .le. zero * abs ( iworker(ix,iy) + oldpixval )
         else if (method .eq. 'boundary') then
           lpixval = abs ( iworker(ix,iy) - boundpixval ) .gt. zero * abs ( iworker(ix,iy) + oldpixval )
         else if (method .eq. 'positive') then 
           lpixval = iworker(ix,iy) .gt. zero .and. .not.lntakenbyanother
         else if (method .eq. 'negative') then
           lpixval = iworker(ix,iy) .le. zero
         else
           write (*,'(/a)') '   LPIXVAL: Invalid method: '//method
           stop 99
         endif
       else
         write (*,'(/a)') '   LPIXVAL: Invalid clogic: '//clogic
         stop 99
       endif
       
       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine push 
       
     &            ( ix, iy, ixstack, iystack, nstack, maxstack )
!__________________________________________________________________________________________________________________________________
!
! Alexander Men'shchikov, DAp IRFU CEA Saclay, France, on March 12, 2009
!__________________________________________________________________________________________________________________________________
!                     
       implicit      none
       integer       ix, iy, nstack, maxstack
       integer       ixstack(maxstack), iystack(maxstack)
!__________________________________________________________________________________________________________________________________
!
       nstack = nstack + 1
       if (nstack .gt. maxstack) then
         write (*,'(/a,i10)') '   PUSH: Stack overflow; current maximum stack size:', maxstack
         stop 99
       endif       
       ixstack(nstack) = ix
       iystack(nstack) = iy
       
       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine pop 

     &            ( ixstack, iystack, nstack, maxstack, ix, iy )
!__________________________________________________________________________________________________________________________________
!
! Alexander Men'shchikov, DAp IRFU CEA Saclay, France, on March 12, 2009
!__________________________________________________________________________________________________________________________________
!                     
       implicit      none
       integer       ix, iy, nstack, maxstack
       integer       ixstack(maxstack), iystack(maxstack)
!__________________________________________________________________________________________________________________________________
!
       ix = ixstack(nstack)
       iy = iystack(nstack)
       nstack = nstack - 1
       
       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       SUBROUTINE fit
      
     &            ( x, y, ndata, sig, mwt, a, b, siga, sigb, chi2, q )
!__________________________________________________________________________________________________________________________________
!
! From Numerical Recipes for FORTRAN.
!
! Given a set of data points x(1:ndata), y(1:ndata) with individual standard deviations sig(1:ndata), fit them to a straight line
! y = a + b * x by minimizing chi^2. Returned are a, b and their respective probable uncertainties siga and sigb, the chi-square
! chi2, and the goodness-of-fit probability q (that the fit would have chi^2 this large or larger). If mwt = 0 on input, then the
! standard deviations are assumed to be unavailable: q is returned as 1.0 and normalization of chi2 is to unit standard deviation
! on all points.
!__________________________________________________________________________________________________________________________________
!                     
       implicit      none
       INTEGER       mwt, ndata, i
       REAL*8        a, b, chi2, q, siga, sigb, sig(ndata), x(ndata), y(ndata)
CU     USES          gammq
       REAL*8        sigdat, ss, st2, sx, sxoss, sy, t, wt, gammq
!__________________________________________________________________________________________________________________________________
!
       sx = 0.0d0
       sy = 0.0d0
       st2 = 0.0d0
       b = 0.0d0
       if (mwt .ne. 0) then
         ss = 0.0d0
         do 11 i=1,ndata
           wt = 1.0d0 / (sig(i)**2)
           ss = ss + wt
           sx = sx + x(i)*wt
           sy = sy + y(i)*wt
11       continue
       else
         do 12 i=1,ndata
           sx = sx + x(i)
           sy = sy + y(i)
12       continue
         ss = float ( ndata )
       endif
       sxoss = sx / ss
       if (mwt .ne. 0) then
         do 13 i=1,ndata
           t = (x(i) - sxoss) / sig(i)
           st2 = st2 + t*t
           b = b + t * y(i) / sig(i)
13       continue
       else
         do 14 i=1,ndata
           t = x(i) - sxoss
           st2 = st2 + t*t
           b = b + t * y(i)
14       continue
       endif
       b = b / st2
       a = (sy - sx * b) / ss
       siga = sqrt ( (1.0d0 + sx * sx / (ss * st2)) / ss)
       sigb = sqrt ( 1.0d0 / st2 )
       chi2 = 0.0d0
       if (mwt .eq. 0) then
         do 15 i=1,ndata
           chi2 = chi2 + (y(i) - a - b * x(i))**2
15       continue
         q = 1.0d0
         sigdat = sqrt ( chi2 / (ndata - 2) )
         siga = siga * sigdat
         sigb = sigb * sigdat
       else
         do 16 i=1,ndata
           chi2 = chi2 + ((y(i) - a - b * x(i)) / sig(i))**2
16       continue
         q = gammq ( 0.5d0 * (ndata - 2), 0.5d0 * chi2 )
       endif

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       function gammq

     &          ( a, x )
!__________________________________________________________________________________________________________________________________
!
! From Numerical Recipes for FORTRAN.
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       real*8        a, gammq, x
CU     uses          gcf, gser
       real*8        gammcf, gamser, gln
!__________________________________________________________________________________________________________________________________
!
       if (x .lt. 0.0d0 .or. a .le. 0.0d0) then
         write (*,*) ' bad arguments in gammq'
         stop
       endif
       if (x .lt. a + 1.0d0)then
         call gser ( gamser, a, x, gln )
         gammq = 1.d0 - gamser
       else
         call gcf ( gammcf, a, x, gln )
         gammq = gammcf
       endif

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       subroutine gcf
       
     &            ( gammcf, a, x, gln )
!__________________________________________________________________________________________________________________________________
!
! From Numerical Recipes for FORTRAN.
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       integer       itmax, i
       real*8        a, gammcf, gln, x, EPS, FPMIN
       parameter   ( itmax = 100, eps = 3.0d-7, fpmin = 1.0d-30)
CU     uses          gammln
       real*8        an, b, c, d, del, h, gammln
!__________________________________________________________________________________________________________________________________
!
       gln = gammln(a)
       b = x + 1.0d0 - a
       c = 1.0d0 / FPMIN
       d = 1.0d0 / b
       h = d
       do 11 i=1,ITMAX
         an = -i * (i - a)
         b = b + 2.0d0
         d = an * d + b
         if (abs ( d ) .lt. FPMIN) d = FPMIN
         c = b + an / c
         if (abs ( c ) .lt. FPMIN) c = FPMIN
         d = 1.0d0 / d
         del = d * c
         h = h * del
         if (abs ( del - 1.0d0 ) .lt. EPS) goto 1
11     continue
       write (*,*) ' a too large, ITMAX too small in gcf'
       stop

1      gammcf = exp ( -x + a * log ( x ) - gln ) * h
     
       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       SUBROUTINE gser
      
     &            ( gamser, a, x, gln )
!__________________________________________________________________________________________________________________________________
!
! From Numerical Recipes for FORTRAN.
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       integer       itmax, n
       real*8        a, gamser, gln, x, EPS
       parameter   ( ITMAX = 100, EPS = 3.0d-7)
CU     uses gammln
       real*8        ap, del, sum, gammln
!__________________________________________________________________________________________________________________________________
!
       gln = gammln ( a )
       if (x .le. 0.0d0)then
         if (x .lt. 0.0d0) then
           write (*,*) ' x < 0 in gser'
           stop
         endif
         gamser = 0.0d0
         return
       endif
       ap = a
       sum = 1.0d0 / a
       del = sum
       do 11 n=1,ITMAX
         ap = ap + 1.0d0
         del = del * x / ap
         sum = sum + del
         if (abs ( del ) .lt. abs ( sum ) * EPS) goto 1
11     continue
       write (*,*) ' a too large, ITMAX too small in gser'
       stop

1      gamser = sum * exp ( -x + a * log ( x ) - gln )

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

       FUNCTION gammln
    
     &         ( xx )
!__________________________________________________________________________________________________________________________________
!
! From Numerical Recipes for FORTRAN.
!__________________________________________________________________________________________________________________________________
!
       implicit      none
       REAL*8        gammln, xx
       INTEGER       j
       REAL*8        ser, stp, tmp, x, y, cof(6)
       SAVE          cof, stp
       DATA cof, stp / 76.18009172947146d0,-86.50532032941677d0,24.01409824083091d0,-1.231739572450155d0,.1208650973866179d-2,
     &                -.5395239384953d-5,2.5066282746310005d0 /
!__________________________________________________________________________________________________________________________________
!
       x = xx
       y = x
       tmp = x + 5.5d0
       tmp = (x + 0.5d0) * log ( tmp ) - tmp
       ser = 1.000000000190015d0
       do 11 j=1,6
         y = y + 1.0d0
         ser = ser + cof(j) / y
11     continue
       gammln = tmp + log ( stp * ser / x )

       return
       end

!||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
